/**CFile***********************************************************************

  FileName    [bmcCmd.c]

  PackageName [bmc]

  Synopsis    [BMC commands.]

  Description [Command file]

  SeeAlso     []

  Author      [Roberto Cavada]

  Copyright   [
  This file is part of the ``bmc'' package of NuSMV version 2. 
  Copyright (C) 2000-2001 by ITC-irst and University of Trento. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "bmcInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

#define BMC_INTERNAL_ERROR ((node_ptr)-1)
#define BMC_NO_TRACE ((node_ptr)0)

#undef  BMC_USE_SATO
#undef  BMC_DISABLE_CNF_DUMP

#define DIMAX_FILENAME_MAXLEN 4096

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/


/* ========================================================================= */
/* This variable is existing only temporary. It is used to store the idx
   of the currently processed property. Access to it using its wrapper. */
#define BMC_NO_PROPERTY -1
static int bmc_GL_idxProperty = BMC_NO_PROPERTY;

static void bmcSetCurrentPropertyIndex(int idx) 
{ 
  nusmv_assert(idx == BMC_NO_PROPERTY || (idx>=0 && idx < Prop_Db_GetSize())); 
  bmc_GL_idxProperty = idx;
} 

static int bmcGetCurrentPropertyIndex() 
{
  /* returns BMC_NO_PROPERTY if no property is currently processed */
  return bmc_GL_idxProperty;
}
/* ========================================================================= */



/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**Macro***********************************************************************

  Synopsis     [Declares a set of variables used by the options saving code]

  Description  [In command processing functions use this macro to declare all
  variables needed to securly save all options might be modified by command
  options, but that you do not want any side-effect on those options value]

  SideEffects  []

  SeeAlso      [BMC_CMD_SAVE_OPTIONS, BMC_CMD_RESTORE_OPTIONS]

******************************************************************************/
#define BMC_CMD_DECL_SAVING_OPTIONS() \
  int curr_k, curr_l;           /* to keep current values of k and l */ \
  char* szCurrDimacsFilename      = NIL(char); \
  char* szCurrInvarDimacsFilename = NIL(char)


/**Macro***********************************************************************

  Synopsis     [Use this macro in command processing functions in order to
  save options default values, so avoiding any side effect on those options]

  Description  [Only one shot is allowed for this macro call. Do not call more
  than one time! Use BMC_CMD_RESTORE_OPTIONS before exit from any 
  function branch which has carried out the BMC_CMD_SAVE_OPTIONS macro]

  SideEffects  [Variables delcared using BMC_CMD_DECL_SAVING_OPTIONS will 
  change]

  SeeAlso      [BMC_CMD_DECL_SAVING_OPTIONS, BMC_CMD_RESTORE_OPTIONS]

******************************************************************************/
#define BMC_CMD_SAVE_OPTIONS() \
  bmcSave_k_l(&curr_k, &curr_l); \
  bmcSaveDimacsFilenames(&szCurrDimacsFilename, &szCurrInvarDimacsFilename)


/**Macro***********************************************************************

  Synopsis     [Exits from the function after have restored all options 
  values saved using BMC_CMD_SAVE_OPTIONS.]

  Description  [Make sure that all function branches containing a call to 
  BMC_CMD_SAVE_OPTIONS close their execution with a call to the 
  BMC_CMD_RESTORE_OPTIONS macro. ]

  SideEffects  [Global options will receive the previuosly saved value]

  SeeAlso      [BMC_CMD_DECL_SAVING_OPTIONS, BMC_CMD_SAVE_OPTIONS]

******************************************************************************/
#define BMC_CMD_RESTORE_OPTIONS() \
  bmcRestore_k_l(curr_k, curr_l);  \
  bmcRestoreDimacsFilenames(szCurrDimacsFilename, szCurrInvarDimacsFilename)
  


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static int UsageBmcSetup    ARGS((void));
static int UsageBmcSimulate ARGS((void));
static int UsageBmcLtlsat   ARGS((void));

static int UsageBmcCheckLtlSpec       ARGS((void));
static int UsageBmcCheckLtlSpecOnePb  ARGS((void));
static int UsageBmcGenLtlSpec         ARGS((void));
static int UsageBmcGenLtlSpecOnePb    ARGS((void));
static int UsageBmcCheckInvar         ARGS((void));
static int UsageBmcGenInvar           ARGS((void));

static outcome bmcSetLoopbackAsString ARGS((const char* str));

static outcome 
bmc_CommandsWith_k_l_OptionsHandling ARGS((int argc, char ** argv, 
					   Prop_Ptr* pProperty,
					   Prop_Type type, 
					   int* dimacsSpecified));
static outcome 
bmc_CommandsWithout_k_l_OptionsHandling ARGS((int argc, char ** argv, 
					      Prop_Ptr* pProperty,
					      Prop_Type type,
					      int* dimacsSpecified));


static void bmcSave_k_l ARGS((int* k_store, int* l_store));
static void bmcRestore_k_l ARGS((const int k_store, const int l_store));
static void bmcSaveDimacsFilenames ARGS((char** szFilenameDimacs, 
					 char** szFilenameInvarDimacs));
static void bmcRestoreDimacsFilenames ARGS((char* szFilenameDimacs, 
					    char* szFilenameInvarDimacs));

static outcome bmcCheck_k_l ARGS((void));

/* error messages: */
static void bmcErrorInvalidProperty ARGS((const char* szFormula));
static void bmcErrorInvalidNumber ARGS((const char* szNumber));
static void bmcErrorInvalidLoopback ARGS((const char* szLoopbackString)); 
static void bmcErrorInvalid_k_l ARGS((const int k, const int l));
static void bmcErrorPropertyAlreadySpecified ARGS((void));

static Fsm_RbcPtr Prop_GetRbcFsmCoi ARGS((Prop_Ptr prop));

static void BmcGenProblem ARGS((rbc_ptr prob,
			       lsList CnfVars,   /*The list of CNF variables*/
			       lsList CnfClauses, /*The list of CNF clauses*/
			       int * CnfMaxVar, /*The max CNF variable index*/
			       int k, int l,
			       int mode));

static void BmcSatGenProblemKL ARGS((rbc_ptr prob, int k, int l));
static void BmcSatGenProblemK  ARGS((rbc_ptr prob, int k));
static void BmcSatGenInvar     ARGS((rbc_ptr prob));

static sat_result BmcSatSolveProblemKL ARGS((rbc_ptr prob, int k, int l, 
					     int bCreateDimacs, 
					     node_ptr* model));
static sat_result BmcSatSolveProblemK  ARGS((rbc_ptr prob, int k, 
					     int bCreateDimacs, 
					     node_ptr* model));

static sat_result BmcSatSolveProblemUptoK ARGS((rbc_ptr prob, int k,
						int bCreateDimacs, 
						node_ptr* model));

static sat_result BmcSatSolveFalsificationProblem ARGS((rbc_ptr prob, 
							int k, int l, int mode,
							int bCreateDimacs,
							node_ptr* model));

static sat_result BmcSatSolveInvar ARGS((rbc_ptr prob, int bCreateDimacs, 
					 node_ptr* model));

static node_ptr BmcPropModel2SMVModel ARGS((lsList SimResult, int CnfMaxVar, 
					    int k));

static void 
BmcExpandMacrosInFilename(const char* filename, char* filename_expanded, 
			  size_t buf_len);

/**AutomaticEnd***************************************************************/

/* prototypes of the command functions */
static int CommandBmcSetup    ARGS((int argc, char ** argv));
static int CommandBmcSimulate ARGS((int argc, char ** argv));
static int CommandBmcLtlsat   ARGS((int argc, char ** argv));
static int CommandCheckLtlSpecBmc      ARGS((int argc, char ** argv));
static int CommandCheckLtlSpecBmcOnePb ARGS((int argc, char ** argv));
static int CommandGenLtlSpecBmc        ARGS((int argc, char ** argv));
static int CommandGenLtlSpecBmcOnePb   ARGS((int argc, char ** argv));
static int CommandGenInvarBmc          ARGS((int argc, char ** argv));
static int CommandCheckInvarBmc        ARGS((int argc, char ** argv));


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Adds all bmc-related commands to the interactive shell]

  Description        []

  SideEffects        []

  SeeAlso            [SmInit]

******************************************************************************/
void
Bmc_AddCmd(void) 
{
  Cmd_CommandAdd("bmc_setup",         CommandBmcSetup, 0);
  Cmd_CommandAdd("bmc_simulate",      CommandBmcSimulate, 0);
  Cmd_CommandAdd("_bmc_ltlsat",       CommandBmcLtlsat, 0);
  Cmd_CommandAdd("check_ltlspec_bmc", CommandCheckLtlSpecBmc, 0); 
  Cmd_CommandAdd("check_ltlspec_bmc_onepb", 
		 CommandCheckLtlSpecBmcOnePb, 0);
  Cmd_CommandAdd("gen_ltlspec_bmc", 
		 CommandGenLtlSpecBmc, 0); 
  Cmd_CommandAdd("gen_ltlspec_bmc_onepb", 
		 CommandGenLtlSpecBmcOnePb, 0); 
  Cmd_CommandAdd("gen_invar_bmc",     CommandGenInvarBmc, 0); 
  Cmd_CommandAdd("check_invar_bmc",   CommandCheckInvarBmc, 0); 
  Cmd_CommandAdd("_bmc_test_tableau", Bmc_TestTableau, 0); 
}

/* ////////////////////////////////////////////////////////////////////
   ////////////////////   MODEL BUILDING COMMANDS    //////////////////
  ///////////////////////////////////////////////////////////////////// */

/**Function********************************************************************

  Synopsis           [Initializes the bmc sub-system, and builds the model in
RBC format]

  Description        []

  SideEffects        [Overall the bmc system]

  SeeAlso            []

  CommandName        [bmc_setup] 	   

  CommandSynopsis    [Builds the model in RBC format.]  

  CommandArguments   [\[-h\]]  

  CommandDescription [You must call this command before use any other 
bmc-related command. Only one call per session is required.]  

******************************************************************************/
static int CommandBmcSetup(int argc, char ** argv)
{
  /* process command options */ 
  {
    int c;

    util_getopt_reset();
    while ((c = util_getopt (argc, argv, "h")) != EOF) {
      switch (c) {
      case 'h': return UsageBmcSetup ();
      }
    }
    if (argc != 1) return UsageBmcSetup ();
  }

  if (cmp_struct_get_read_model(cmps) == 0) {
    fprintf(nusmv_stderr,
            "A model must be read before. Use the \"read_model\" command.\n");
    return(1);
  }

  if (cmp_struct_get_flatten_hrc(cmps) == 0) {
    fprintf(nusmv_stderr,
            "The hierarchy must be flattened before. Use the \"flatten_hierarchy\" command.\n");
    return(1);
  }

  if (cmp_struct_get_encode_variables(cmps) == 0) {
    fprintf(nusmv_stderr, "The variables must be built before. Use the \"encode_variables\" command.\n");
    return(1);
  }

  if (cmp_struct_get_build_bool_model(cmps) == 0) {
    fprintf(nusmv_stderr, "The model must be built before.\n");
    if (opt_verbose_level_gt(options, 1)) {
      fprintf(nusmv_stderr, "Use the \"build_model\" command.\n");
    }
    return(1);
  }

  if (cmp_struct_get_bmc_setup(cmps)) {
    fprintf (nusmv_stderr, "A call to bmc_setup has already been done.\n");
    return(1);
  }

  Bmc_Setup();

  cmp_struct_set_bmc_setup(cmps);
  return(0);
}

void Bmc_Setup()
{
    Fsm_SexpPtr bfsm = Prop_MasterGetBoolFsm();
    node_ptr init_bexp = Compile_FsmSexpGetInit(bfsm);
    node_ptr trans_bexp = Compile_FsmSexpGetTrans(bfsm);
    node_ptr invar_bexp = Compile_FsmSexpGetInvar(bfsm);
    node_ptr fairness_bexp = Compile_FsmSexpGetFairness(bfsm);
    Fsm_RbcPtr rbc_fsm = Compile_FsmRbcAlloc(); /* Create empty RBC FSM */

    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Initializing BMC module ... \n");
    }
    Bmc_Init();

    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Constructing RBC for init ... \n");
    }
    Compile_FsmRbcSetInit(rbc_fsm, Bmc_Bexp2Rbc(init_bexp));
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Constructing RBC for trans ... \n");
    }
    Compile_FsmRbcSetTrans(rbc_fsm, Bmc_Bexp2Rbc(trans_bexp));

    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Constructing RBC for invar ... \n");
    }
    Compile_FsmRbcSetInvar(rbc_fsm, Bmc_Bexp2Rbc(invar_bexp));

    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Constructing RBC for fairness ... \n");
    }
    {
      node_ptr fairness_rbc; 
      node_ptr list_valid_fairness;

      /* check that all fairnesses are strictly in propositional form. 
       list_valid_fairness is the list of valid wff only: */
      list_valid_fairness = Bmc_CheckFairnessListForPropositionalFormulae
	(fairness_bexp);
      
      fairness_rbc = Bmc_MapBexp2Rbc(list_valid_fairness);
      Compile_FsmRbcSetFairness(rbc_fsm, fairness_rbc);

      free_list(list_valid_fairness);
    }

    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Storing into FSM ... \n");
    }
    Prop_MasterSetRbcFsm(rbc_fsm);
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "done!\n");
    }
}


/**Function********************************************************************

  Synopsis           [Usage string for CommandBmcSetup]

  Description        []

  SideEffects        []

  SeeAlso            [CommandBmcSetup]
******************************************************************************/
static int UsageBmcSetup (void)
{
  fprintf (nusmv_stderr, "usage: bmc_setup [-h]\n");
  fprintf (nusmv_stderr, "  -h \t\tPrints the command usage.\n");
  fprintf (nusmv_stderr, "   \t\tGenerates the RBC-based representation of the model.\n");

  return(1);
}


/**Function********************************************************************

  Synopsis           [CommandBmcSimulate generates a trace of the problem 
represented from the simple path from 0 (zero) to k]

  Description        [CommandBmcSimulate does not require a specification 
to build the problem, because only the model is used to build it.]

  SideEffects        [None]

  SeeAlso            []

  CommandName        [bmc_simulate] 	   

  CommandSynopsis    [Generates a trace of the model from 0 (zero) to k]  

  CommandArguments   [\[-h\]]  

  CommandDescription [bmc_simulate does not require a specification 
to build the problem, because only the model is used to build it.]  

******************************************************************************/
static int CommandBmcSimulate(int argc, char ** argv)
{
  BMC_CMD_DECL_SAVING_OPTIONS();

  BMC_CMD_SAVE_OPTIONS();
  {  /* process command options */ 
    int c;
    util_getopt_reset();
    while((c = util_getopt(argc, argv, "hk:")) != EOF) {
      switch (c) {
      case 'h':
	BMC_CMD_RESTORE_OPTIONS(); 
	return UsageBmcSimulate();

      case 'k':
	{
	  char *strNumber = util_strsav(util_optarg);
	  int k;
	  if (util_str2int(strNumber, &k) != 0) {
	    bmcErrorInvalidNumber(strNumber);
	    BMC_CMD_RESTORE_OPTIONS(); 
	    return 1;
	  }
	  
	  /* k and l constrains are going to be check out of switch */ 
	  set_bmc_pb_length(options, k); 
	  break;  
	}
      } /* switch stmt */
    } /* while */ 
    if (argc != util_optind) {
      BMC_CMD_RESTORE_OPTIONS(); 
      return UsageBmcSimulate();
    }
  
    /* ----------------------------------------------------------------------*/
    /* Checking of k,l constrains:                                           */
    if (bmcCheck_k_l() != SUCCESS) {
      bmcErrorInvalid_k_l(get_bmc_pb_length(options), 
			  get_bmc_pb_loop_user_pov(options));
      BMC_CMD_RESTORE_OPTIONS(); 
      return 1;
    }
  } /* option handling */
  
  /* make sure the model hierarchy has been flattened */
  if (cmp_struct_get_flatten_hrc (cmps) == 0) {
    fprintf (nusmv_stderr, "The hierarchy must be flattened before. Use "\
	     "the \"flatten_hierarchy\" command.\n");
    
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }

  /* make sure bmc has been set up */
  if (cmp_struct_get_bmc_setup(cmps) == 0) {
    fprintf (nusmv_stderr, "Bmc must be setup before. Use "\
	     "the \"bmc_setup\" command.\n");
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }

  {
    Fsm_RbcPtr rbc_fsm;
    int k;
    rbc_ptr prob; /* The problem in RBC format */
    node_ptr trace;
    sat_result satRes;

    rbc_fsm = Prop_MasterGetRbcFsm();

    k = get_bmc_pb_length (options);
    
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, 
	      "\nGenerating simulation trace of length %d (no loopback)\n",
	      k);	
    }

    prob = Bmc_Mk_k_Path(rbc_fsm, k);

    satRes = BmcSatSolveProblemK(prob, k, FALSE, &trace);

    if (satRes == SAT_INTERNAL_ERROR) {
      /* internal error */
      nusmv_assert(false);
    }
    else if (satRes == SAT_UNSATISFIABLE_PROBLEM) {
      fprintf(nusmv_stdout, 
	      "The model deadlocks before requested length %d!\n", k);
      BMC_CMD_RESTORE_OPTIONS(); 
      return 1;
    }
    else {
      /* no other choices are supplied: */
      nusmv_assert(satRes==SAT_SATISFIABLE_PROBLEM);

      fprintf(nusmv_stderr, "\n\n");
      print_explanation(trace, all_symbols);

      free_list(trace); /* destroys the unnecessary list */
    }
  }
  BMC_CMD_RESTORE_OPTIONS(); 
  return 0;
}


/**Function********************************************************************

  Synopsis           [Usage string for UsageBmcSimulate]

  Description        []

  SideEffects        [None]

  SeeAlso            [CommandBmcSetup]

******************************************************************************/
static int UsageBmcSimulate (void)
{
  fprintf (nusmv_stderr, "usage: bmc_simulate [-h]\n");
  fprintf (nusmv_stderr, "  -h \t\tPrints the command usage.\n");
  fprintf (nusmv_stderr, "   \t\tGenerates a k-steps simulation using Bounded Model Checking.\n" \
	   "\t\tYou can set k setting the variable bmc_length.\n");

  return(1);
}



/**Function********************************************************************

  Synopsis           [Builds a problem for given k and l, then 
  calls the solver]

  Description        []

  SideEffects        []

  SeeAlso            [BmcSatSolveProblemKL]

  CommandName        [_bmc_ltlsat] 	   

  CommandSynopsis    [Builds a problem for given k and l, then 
  calls the solver]  

  CommandArguments   [\[-h | -n <i>ltl_property_index</i> | "<ltl spec>"\]]  

  CommandDescription [
  This command checks that the given property index matched to a valid LTL 
  formula. k and l must be specified using the bmc_length and bmc_loopback 
  envirnoment variables. 
  <p>
  Command options:<p>
  <dl>
    <dt> <tt>-n <i>ltl_property_index</i></tt>
       <dd> <i>ltl_property_index</i> is the index of a valid LTL specification formula used to build the problem. <BR>
The validity of <i>ltl_property_index</i> value is checked out by the system.
    <dt> <tt>"<i>ltl spec</i>"</tt>
       <dd> This command is able to parse and check a directly specified LTL formula.
  </dl>]

******************************************************************************/
static int CommandBmcLtlsat(int argc, char ** argv)
{
  Prop_Ptr ltlprop=NIL_PTR(Prop_Ptr);   /* The property being processed */

  {
    int c;
    char *strNumber = NIL(char);
    int prop_no;

    util_getopt_reset();
    while((c = util_getopt(argc, argv, "hn:")) != EOF) {
      switch (c) {
      case 'h':
	return UsageBmcLtlsat();
      case 'n':
	{
	  if (strNumber != NIL(char)) return(UsageBmcLtlsat());
	  strNumber = util_strsav(util_optarg);
	  {
	    char *err_occ[2];

	    err_occ[0] = "";
	    prop_no = strtol(strNumber, err_occ, 10);
	    if ((strcmp(err_occ[0], "") != 0)) {
	      fprintf(nusmv_stderr,
		      "Error: \"%s\" is not a valid value (must be integer).\n", strNumber);
	      return(1);
	    }
	  }
	  if (prop_no >= Prop_Db_GetSize() || prop_no < 0) {
	    fprintf(nusmv_stderr,
		    "Error: \"%s\" is not a valid value, must be in the range [0,%d].\n", \
		    strNumber, Prop_Db_GetSize()-1);
	    return(1);
	  }
	  ltlprop = Prop_Db_GetNum(prop_no);
	  if (Prop_GetType(ltlprop) != Prop_Ltl) {
	    fprintf(nusmv_stderr,
		    "Error: property %d is not of type LTL\n", prop_no);
	    return(1);
	  }
	  break;
	}
      }
    }
    if (argc != util_optind) return(UsageBmcLtlsat ());
  }


  /* make sure the model hierarchy has been flattened */
  if (cmp_struct_get_flatten_hrc (cmps) == 0) {
    fprintf (nusmv_stderr, "The hierarchy must be flattened before. Use "\
	     "the \"flatten_hierarchy\" command.\n");
    
    return(1);
  }

  /* make sure bmc has been set up */
  if (cmp_struct_get_bmc_setup(cmps) == 0) {
    fprintf (nusmv_stderr, "Bmc must be setup before. Use "\
	     "the \"bmc_setup\" command.\n");
    return(1);
  }

  { /* The problem in RBC format */
    node_ptr ltlspec;
    node_ptr fltlspec;
    node_ptr bltlspec;
    int k;
    int l;
    rbc_ptr prob;
    node_ptr trace = LS_NIL;
    sat_result satRes;

    /*If no property index had been specified, try to read from command line:*/
    if (ltlprop == NIL_PTR(Prop_Ptr)) {
      int result;
      node_ptr cmd_ltl_formula = NIL_PTR(node_ptr);
      
      result = Parser_ReadCmdFromString(argc, argv, "LTLWFF ", ";\n", 
					&cmd_ltl_formula);
      if ( (result == 0) && (cmd_ltl_formula!= NIL_PTR(node_ptr)) ) {
	if (node_get_type(cmd_ltl_formula) == LTLWFF) {
	  CATCH 
	    {
	      int pindex = Prop_Db_PropCreateAndAdd( car(cmd_ltl_formula), 
						     Prop_Ltl );
	      if (pindex != -1) {
		/* assign here the added property: */
		ltlprop = Prop_Db_GetNum(pindex);
		//nusmv_assert(Prop_GetType(ltlprop) != Prop_Ltl);
	      }
	    } /* end CATCH */
	}
	else {
	  fprintf(nusmv_stderr, 
		  "Parsing error: expected \"_bmc_ltlsat ltl_expr [IN context]\", where ltl_expr is a valid LTL formula.\n");
	  return 1;
	}
      }
    } 

    /* checks that a property was selected: */
    if (ltlprop == (Prop_Ptr)NULL) {
      fprintf(nusmv_stderr, 
	      "A LTL property must be specified\n");
      return 1;
    }

    /* set global property idx (for filename pattern expansion): */
    bmcSetCurrentPropertyIndex(Prop_GetIndex(ltlprop));


    ltlspec = Prop_GetProp(ltlprop);
    fltlspec = Compile_FlattenSexpExpandDefine(ltlspec, Nil);
    
    bltlspec = detexpr2bexpr(fltlspec);
    bltlspec = Bmc_LtlWffMkNnfToplevel(bltlspec);

    k = get_bmc_pb_length (options);
    l = get_bmc_pb_loop (options);


    if (l == BMC_ALL_LOOPS) {
      /* not implemented yet */
      fprintf (nusmv_stderr, "Error: the case 'all loops' is not implemented yet.\nPlease set the variable 'bmc_loopback' to another value.\n\n");

      /* reset currently processed property index: */
      bmcSetCurrentPropertyIndex(BMC_NO_PROPERTY);
      return(1);
    }

    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "\nLTL tableau with bound %d, ", k);
      if (l == BMC_NO_LOOP) {
	fprintf(nusmv_stderr, "no loopback\n");
      } else {
	fprintf(nusmv_stderr, "loopback at %d\n", l);
      }
    }

    prob = Bmc_MkTableau_k_l(bltlspec, k, l);

    satRes = BmcSatSolveProblemKL(prob, k, l, FALSE, &trace);

    if (satRes == SAT_INTERNAL_ERROR) {
      /*  internal error */
      nusmv_assert(false);
    }
    else if (satRes == SAT_UNSATISFIABLE_PROBLEM) {
      fprintf(nusmv_stdout, "-- the ");
      print_spec(nusmv_stdout, ltlspec); 
      fprintf(nusmv_stdout, " is not satisfiable at length %d\n", k);
    }
    else {
      /* no other choices are supplied: */
      nusmv_assert(satRes==SAT_SATISFIABLE_PROBLEM);

      fprintf(nusmv_stdout, "-- the ");
      print_spec(nusmv_stdout, ltlspec); 
      fprintf(nusmv_stdout, " is satisfiable at length %d\n", k);
      print_explanation(trace, all_symbols);
      
      free_list(trace); /* destroys the unnecessary list */
    }
  }

  /* reset currently processed property index: */
  bmcSetCurrentPropertyIndex(BMC_NO_PROPERTY);
  return(0);
}

/**Function********************************************************************

   Synopsis           [Usage string for UsageBmcLtlsat]

  Description        []

  SideEffects        [None]

  SeeAlso            [bmc_CommandBmcPbGen]

******************************************************************************/
static int UsageBmcLtlsat (void)
{
  fprintf (nusmv_stderr, 
	   "usage: _bmc_ltlsat -h | -n <p> | \"<ltl property>\"\n");
  fprintf (nusmv_stderr, "  -h \t\tPrints the command usage.\n");
  fprintf (nusmv_stderr,
 "  -n \t\tChecks the k-steps satisfiability of LTL property of index <p>.\n  \"<ltl property>\" you can check the specified formula, which will be added to the properties database.\n\t\tYou can set k setting the variable bmc_length.\n");

  return(1);
}

/**Function********************************************************************

  Synopsis           [Genetate DIMACS version and/or solve and INVARSPEC 
                      problem.]

  Description        []

  SideEffects        []

  SeeAlso            [Bmc_GenSolvePbs]

******************************************************************************/
void Bmc_GenSolveInvar(Prop_Ptr invarprop,  const int bCreateDimacs, 
		       const int bSolve) 
{
  node_ptr invarspec;   /* The corresponding wff */
  node_ptr finvarspec;  /* wff after flattening*/
  node_ptr binvarspec;  /* Its booleanization */
  Fsm_RbcPtr rbc_fsm;   /* The corresponding rbc fsm  */
  rbc_ptr prob; /* The problem in RBC format */

  /* checks that a property was selected: */
  nusmv_assert(invarprop != (Prop_Ptr)NULL);


  /* checks if it has been already checked: */
  if (Prop_GetStatus(invarprop) != Prop_Unchecked) {
    /* aborts this check */
    return;
  }
  
  /* set global property idx (for filename pattern expansion): */
  bmcSetCurrentPropertyIndex(Prop_GetIndex(invarprop));

  invarspec = Prop_GetProp(invarprop);
  finvarspec = Compile_FlattenSexpExpandDefine(invarspec, Nil);
  binvarspec = detexpr2bexpr(finvarspec);
  binvarspec = Bmc_LtlWffMkNnfToplevel(binvarspec);
  rbc_fsm = Prop_GetRbcFsmCoi(invarprop);

  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "\nGenerating invariant problem\n");
  }

  prob = Bmc_MkInvariantProblem(rbc_fsm, binvarspec);

  if (bSolve == FALSE)  BmcSatGenInvar(prob);
  else {
    node_ptr trace = LS_NIL;
    sat_result satRes;  

    /* Dump to dimacs if required, and call also the solver: */ 
    satRes = BmcSatSolveInvar(prob, bCreateDimacs, &trace);

    /* --------------------------------------------------------------------- */
    /* Process SAT results: */
    if (satRes == SAT_INTERNAL_ERROR) {
      /* internal error */
      free_list(trace); /* destroys the unnecessary list */
      internal_error("Sorry, solver answered with a fatal Internal Failure during invar solving.\n");
    }
    else if (satRes == SAT_UNSATISFIABLE_PROBLEM) {
      fprintf(nusmv_stdout, "-- ");
      print_invar(nusmv_stdout, invarspec);
      fprintf(nusmv_stdout, "  is true\n");

      Prop_SetStatus(invarprop, Prop_True);
    }
    else {
      /* no other choices are supplied: */
      nusmv_assert(satRes == SAT_SATISFIABLE_PROBLEM);

      fprintf(nusmv_stdout, "-- cannot prove the ");
      print_invar(nusmv_stdout, invarspec);
      fprintf(nusmv_stdout, ": the induction fails\n");
      print_explanation(trace, all_symbols);
    } 
    /* end of SAT result process */
    /* --------------------------------------------------------------------- */
    
    free_list(trace); /* destroys the unnecessary list */
  } /* call solver */

  /* reset currently processed property index: */
  bmcSetCurrentPropertyIndex(BMC_NO_PROPERTY);
} 


/**Function********************************************************************

  Synopsis           [Given a LTL property generates and solve the problems 
  for all Ki (k_min<=i<=k_max). If bIncreaseK is 0 then k_min==k_max==k and
  only one problem is generated. If bIncreaseK is 1 then k_min == 0 and
  k_max == k. 
  Each problem Ki takes into account of all possible loops from k_min to Ki 
  if loopback is '*' (BMC_ALL_LOOPS). <BR>
  If bCreateDimacs == TRUE then create dimacs before solving. <BR>
  Solver is called only if bSolve == TRUE, otherwise each problem is only 
  generated and dumped to dimacs file]

  Description        [TODO]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void Bmc_GenSolvePbs(Prop_Ptr ltlprop, const int bIncreaseK, 
		     const int bCreateDimacs, const int bSolve) 
{
  node_ptr ltlspec;   /* The corresponding wff */
  node_ptr fltlspec;  /* wff after flattening */
  node_ptr bltlspec;  /* Its booleanization */
  Fsm_RbcPtr rbc_fsm; /* The corrsponding rbc fsm  */

  /* ----------------------------------------------------------------------*/
  /* Here a property was selected                                          */
  /* ----------------------------------------------------------------------*/
  int k_max = get_bmc_pb_length(options);
  int k_min = 0;
  int k_curr = 0; /* to old current value in bmc_length option */
  int increasingK;

  if (bIncreaseK == FALSE) k_min = k_max; 

  /* checks that a property was selected: */
  nusmv_assert(ltlprop != (Prop_Ptr)NULL); 
  

  /* checks if it has been already checked: */
  if (Prop_GetStatus(ltlprop) != Prop_Unchecked) {
    /* aborts this check */
    return;
  }

  /* set global property idx (for filename pattern expansion): */
  bmcSetCurrentPropertyIndex(Prop_GetIndex(ltlprop));

  ltlspec = Prop_GetProp(ltlprop);
  fltlspec = Compile_FlattenSexpExpandDefine(ltlspec, Nil);
  bltlspec = detexpr2bexpr(fltlspec);
  bltlspec = Bmc_LtlWffMkNnfToplevel(Bmc_WffMkNot(bltlspec));
  rbc_fsm = Prop_GetRbcFsmCoi(ltlprop);

  /* save current value in bmc_length: */
  k_curr = get_bmc_pb_length(options);
  nusmv_assert(k_curr >= 0);
    
  /* Start problems generations: */
  for(increasingK = k_min; increasingK <= k_max; ++increasingK) {
    int l; /* problem loopback */
    char szLoop[16]; /* to keep loopback string */
    rbc_ptr prob; /* The problem in RBC format */

    /* set new value for the bmc_length option: */
    set_bmc_pb_length(options, increasingK);

    /* problem loopback might depend from length if loopback is relative: */
    l = get_bmc_pb_loop(options); 
    Bmc_ConvertLoopbackFromInteger(l, szLoop, sizeof(szLoop)); /* to string */

    if (opt_verbose_level_gt(options, 0)) {
      switch(l) {
      case BMC_NO_LOOP:
	fprintf(nusmv_stderr, 
		"\nGenerating problem with bound %d, no loopback...\n", 
		increasingK);
	break;
	
      case BMC_ALL_LOOPS:
	fprintf(nusmv_stderr, 
		"\nGenerating problem with bound %d, all possible loopbacks...\n", 
		increasingK);
	break;

      default:
	/* l can be negative iff loopback from the user pov is < -length */ 
	if (l<=increasingK && l>=0) {	  
	  fprintf(nusmv_stderr, 
		  "\nGenerating problem with bound %d, loopback %s...\n",
		  increasingK, szLoop);
	}
      } /* switch case */
    } /* verbose messages */

    /* checks for loopback vs k compatibility */ 
    if ((l != BMC_NO_LOOP) && (l != BMC_ALL_LOOPS) && 
	((l > increasingK) || (l<0))) {
      fprintf(nusmv_stderr, 
	      "\nWarning: problem with bound %d and loopback %s is not allowed: skipped\n",
	      increasingK, szLoop);
      continue;
    }	  

    if (l == BMC_ALL_LOOPS) {
      prob = Bmc_MkProblem_k(rbc_fsm, bltlspec, increasingK);
    }
    else {
      prob = Bmc_MkProblem_k_l(rbc_fsm, bltlspec, increasingK, l);
    }

    if (bSolve == FALSE) {
      /* Problem generation only: */
      if (l == BMC_ALL_LOOPS) {      
	BmcSatGenProblemK(prob, increasingK);
      }
      else {
	BmcSatGenProblemKL(prob, increasingK, l);
      }      
    }
    else {
      /* SAT invocation */
      node_ptr trace = LS_NIL;
      sat_result satRes;

      if (l == BMC_ALL_LOOPS) {      
	satRes = BmcSatSolveProblemK(prob, increasingK, bCreateDimacs, &trace);
      }
      else {
	satRes = BmcSatSolveProblemKL(prob, increasingK, l, 
				      bCreateDimacs, &trace);
      }

      /* ------------------------------------------------------------------- */
      /* SAT result processing: */
      if (satRes == SAT_INTERNAL_ERROR) {
	/* solver internal error */
	free_list(trace); /* destroys the unnecessary list */
	internal_error("Sorry, solver answered with a fatal Internal Failure during problem solving.\n");
      }
      else if (satRes == SAT_UNSATISFIABLE_PROBLEM) {
	char szLoopMsg[16]; /* for loopback part of message */
	memset(szLoopMsg, 0, sizeof(szLoopMsg));

	if (l == BMC_ALL_LOOPS) {
	  strncpy(szLoopMsg, "", sizeof(szLoopMsg)-1);
	}
	else if (l == BMC_NO_LOOP) {
	  strncpy(szLoopMsg, " and no loop", sizeof(szLoopMsg)-1);
	}
	else {
	  /* loop is Natural: */
	  strncpy(szLoopMsg, " and loop at ", sizeof(szLoopMsg)-1);
	  strncat(szLoopMsg, szLoop, sizeof(szLoopMsg)-1-strlen(szLoopMsg));
	}

	fprintf(nusmv_stdout, 
		"-- no counterexample found with bound %d%s for ", 
		increasingK, szLoopMsg);
	print_spec(nusmv_stdout, ltlspec);
	fprintf(nusmv_stdout, "\n");
      }
      else {
	/* no other choices are supplied :*/
	nusmv_assert(satRes == SAT_SATISFIABLE_PROBLEM);
	
	fprintf(nusmv_stdout, "-- ");
	print_spec(nusmv_stdout, ltlspec);
	fprintf(nusmv_stdout, "  is false\n");
      
	Prop_SetStatus(ltlprop, Prop_False);
	Prop_SetTrace(ltlprop, get_trace_number());
      
	print_explanation(trace, all_symbols);
      }
      /* End of SAT result processing */
      /* ------------------------------------------------------------------- */

      free_list(trace); /* destroys the unnecessary list */
      
      if (satRes == SAT_SATISFIABLE_PROBLEM) {
	break; /* exits from the increasing k loop (found a solution!)*/
      }
    } /* end of SAT invocation */

  } /* for all problems length */


  /* restore old length value: */
  set_bmc_pb_length(options, k_curr);
  
  /* reset currently processed property index: */
  bmcSetCurrentPropertyIndex(BMC_NO_PROPERTY);
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis           []

  Description        [Takes an RBC problem with bound K, generates
  DIMACS, calls SATO, call SIM, and reconstruct model (if any).

  Modes allow to factorize over different file names.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/


#define KLPB_MODE 1
#define KPB_MODE 2
#define UPTOKPB_MODE 3
#define INVARPB_MODE 4

#define NOARG -1000

/**Function********************************************************************

  Synopsis           [Generates the problem taking into account of <B>k</B> 
  and <B>l</B> parameters, and writes the problem into a dimacs file]

  Description        [The generated dimacs file name depends on the model name,
  k, l, and the value of the <B>bmc_prefix</B> environment variable]

  SideEffects        []

  SeeAlso            [BmcGenProblem]

******************************************************************************/
static void
BmcSatGenProblemKL(rbc_ptr prob, int k, int l)
{
  lsList CnfVars = lsCreate();    /* The list of CNF variables */
  lsList CnfClauses = lsCreate(); /* The list of CNF clauses */
  int CnfMaxVar;                  /* The maximum CNF variable index */
  BmcGenProblem(prob, CnfVars, CnfClauses, &CnfMaxVar, k, l, KLPB_MODE);

  /* clean up all unnecessary structures */
  BmcIntFreeListOfLists(CnfClauses);
  lsDestroy(CnfVars, NULL);
}


/**Function********************************************************************

  Synopsis           [Generates the problem taking into account of <B>k</B> 
  parameter, and writes the problem into a dimacs file]

  Description        [The generated dimacs file name depends on the model name,  k, and the value of the <B>bmc_prefix</B> environment variable]

  SideEffects        []

  SeeAlso            [BmcGenProblem]

******************************************************************************/
static void
BmcSatGenProblemK(rbc_ptr prob, int k)
{
  lsList CnfVars = lsCreate();    /* The list of CNF variables */
  lsList CnfClauses = lsCreate(); /* The list of CNF clauses */
  int CnfMaxVar;                  /* The maximum CNF variable index */

  BmcGenProblem(prob, CnfVars, CnfClauses, &CnfMaxVar, k, NOARG, KPB_MODE);

  /* clean up all unnecessary structures */
  BmcIntFreeListOfLists(CnfClauses);
  lsDestroy(CnfVars, NULL);
}


/**Function********************************************************************

  Synopsis           [Generates the invariant problem taking a INVAR 
  specification, and writes the problem into a dimacs file]

  Description        [The generated dimacs file name depends on the model name
  and on the value of the <B>bmc_prefix</B> environment variable]

  SideEffects        []

  SeeAlso            [BmcGenProblem]

******************************************************************************/
static void
BmcSatGenInvar(rbc_ptr prob)
{
  lsList CnfVars = lsCreate();    /* The list of CNF variables */
  lsList CnfClauses = lsCreate(); /* The list of CNF clauses */
  int CnfMaxVar;                  /* The maximum CNF variable index */

  BmcGenProblem(prob, CnfVars, CnfClauses, &CnfMaxVar, 1, NOARG, INVARPB_MODE);

  /* clean up all unnecessary structures */
  BmcIntFreeListOfLists(CnfClauses);
  lsDestroy(CnfVars, NULL);
}


static void
BmcGenProblem(rbc_ptr prob,
	      lsList CnfVars,    /* The list of CNF variables */
	      lsList CnfClauses, /* The list of CNF clauses */
	      int * CnfMaxVar,   /* The maximum CNF variable index */
	      int k,
	      int l,
	      int mode)
{
  char szDimacsName[DIMAX_FILENAME_MAXLEN];

  /*************************************************************
   * RBC TO CNF CONVERSION
   *************************************************************/
  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "\nConverting RBC into CNF...\n");
  }

  (*CnfMaxVar) = Bmc_RbcConvertToCnf(Bmc_VarsMgrGetRbcMgr(),
				 prob, CnfClauses, CnfVars);

  if (opt_verbose_level_gt(options, 1)) {
    fprintf(nusmv_stderr, "Conversion returned CnfMaxVar = %d\n", *CnfMaxVar);
    fprintf(nusmv_stderr, "Length of CnfClauses = %d\n", lsLength(CnfClauses));
    fprintf(nusmv_stderr, "Length of CnfVars = %d\n", lsLength(CnfVars));
  }


  /* selects filename for dimacs output*/
  switch(mode) {
  case KLPB_MODE:
  case KPB_MODE:
  case UPTOKPB_MODE:
    BmcExpandMacrosInFilename(get_bmc_dimacs_filename(options), 
			      szDimacsName,
			      DIMAX_FILENAME_MAXLEN);
				
    break;

  case INVARPB_MODE:
    BmcExpandMacrosInFilename(get_bmc_invar_dimacs_filename(options), 
			      szDimacsName,
			      DIMAX_FILENAME_MAXLEN);
    break;
    
  default:
    /* invalid value for 'mode' */
    nusmv_assert(false);
  } /* switch case */


  /* Dump problem to dimacs file: */
  if ( Bmc_DimacsWriteProblem(szDimacsName, k, CnfClauses, CnfVars, 
			      *(CnfMaxVar)) != SUCCESS ) {
    fprintf(nusmv_stdout, 
"\n*************    WARNING    *************\
 \n An error has occured when writing the file \"%s\".\
 \n DIMACS dumping is not allowed.\
 \n*************  END WARNING  *************\n\n", szDimacsName);
  }
}


static sat_result BmcSatSolveProblemKL(rbc_ptr prob, int k, int l,   
				       int bCreateDimacs, node_ptr* model)
{
  return BmcSatSolveFalsificationProblem(prob, k, l, KLPB_MODE,
					 bCreateDimacs, model);
}

static sat_result BmcSatSolveProblemK(rbc_ptr prob, int k, 
				      int bCreateDimacs, node_ptr* model) 
{
  return BmcSatSolveFalsificationProblem(prob, k, NOARG, KPB_MODE, 
					 bCreateDimacs, model);
}

static sat_result BmcSatSolveProblemUptoK(rbc_ptr prob, int k,  
					  int bCreateDimacs, node_ptr* model) 
{
  return BmcSatSolveFalsificationProblem(prob, k, NOARG, UPTOKPB_MODE, 
					 bCreateDimacs, model);
}


/**Function********************************************************************

  Synopsis           [Given a problem calls the solver and returns a 
  counterexample if there is any]

  Description        [
Converts the rbc-based problem into its CNF representation, and calls the
solver, which can be internal or external, or both. In this last case results
are compared. 
<BR>
Returns:
<ul>
<li> SAT_INTERNAL_ERROR         if an error occurred
<li> SAT_UNSATISFIABLE_PROBLEM  if the formula is unsatisfiable
<li> SAT_SATISFIABLE_PROBLEM    if the assignment satisfying 
the formula. In this case the <i>model</i> parameter points to a list of 
literals
</ul>
]

  SideEffects        []

  SeeAlso            [Bmc_RbcConvertToCnf, Bmc_CallSimSolver, 
  Bmc_CallSatoSolver]

******************************************************************************/
static sat_result 
BmcSatSolveFalsificationProblem(rbc_ptr prob, int k, int l,  
				int mode, int bCreateDimacs, node_ptr* model)
{
  char szDimacsName[DIMAX_FILENAME_MAXLEN];

  lsList CnfVars     = lsCreate(); /* The list of CNF variables */
  lsList CnfClauses  = lsCreate(); /* The list of CNF clauses */
  lsList lsSimResult = lsCreate();
  int CnfMaxVar;                   /* The maximum CNF variable index */

  sat_result SatoResult; /* The result returned by Sato */
  sat_result SimResult;  /* The result returned by Sim */
  sat_result result;     /* used as return value storage room */

  *model = LS_NIL; 

  if (Bmc_RbcIsTrue(Bmc_VarsMgrGetRbcMgr(), prob)) {
    result=SAT_SATISFIABLE_PROBLEM;
    
    if (opt_verbose_level_gt(options, 0)) {    
      fprintf(nusmv_stderr, 
	 "The given problem has been evaluated as the constant value true.\n");
    }
  }
  else if (Bmc_RbcIsFalse(Bmc_VarsMgrGetRbcMgr(), prob)) {
    result=SAT_UNSATISFIABLE_PROBLEM;

    if (opt_verbose_level_gt(options, 0)) {    
      fprintf(nusmv_stdout, 
        "The given problem has been evaluated as the constant value false.\n");
    }
  }
  else {
    /*************************************************************
     * RBC TO CNF CONVERSION
     *************************************************************/
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "\nConverting RBC into CNF...\n");
    }

    CnfMaxVar = Bmc_RbcConvertToCnf(Bmc_VarsMgrGetRbcMgr(), prob, 
	CnfClauses, CnfVars); 

    if (opt_verbose_level_gt(options, 1)) {
      fprintf(nusmv_stderr, "Conversion returned CnfMaxVar = %d\n", CnfMaxVar);
      fprintf(nusmv_stderr, "Length of CnfClauses = %d\n", 
	      lsLength(CnfClauses));
      fprintf(nusmv_stderr, "Length of CnfVars = %d\n", lsLength(CnfVars));
      }

    /* selects filename for dimacs output*/
    switch(mode) {
    case KLPB_MODE:
    case KPB_MODE:
    case UPTOKPB_MODE:
      BmcExpandMacrosInFilename(get_bmc_dimacs_filename(options), 
				szDimacsName,
				DIMAX_FILENAME_MAXLEN);
				
      break;

    case INVARPB_MODE:
      BmcExpandMacrosInFilename(get_bmc_invar_dimacs_filename(options), 
				szDimacsName,
				DIMAX_FILENAME_MAXLEN);
      break;

    default:
      /* invalid value for 'mode' */
      nusmv_assert(false);
    } /* switch case */
    
    /*************************************************************
     * CALL THE SOLVER(S)
     *************************************************************/

#ifdef BMC_USE_SATO
    /*************************************************************
     * CALL SATO AS EXTERNAL SOLVER
     *************************************************************/

    /* dump: */
    if ( Bmc_DimacsWriteProblem(szDimacsName, k, CnfClauses, 
				CnfVars, CnfMaxVar) != SUCCESS ) {
      fprintf(nusmv_stdout, 
"\n*************    WARNING    *************\
 \n An error has occured when writing the file \"%s\".\
 \n DIMACS dumping is not allowed, and SATO cannot be invoked.\
 \n*************  END WARNING  *************\n\n", szDimacsName);
      SatoResult = SAT_UNAVAILABLE;
    }
    else {
      /* invoke the external solver */
      SatoResult = Bmc_CallSatoSolver(szDimacsName);
    }

#else /* not using SATO */
    SatoResult = SAT_UNAVAILABLE; 
#endif

     /*************************************************************
     * CALL SIM SOLVER INTERNALLY 
     * Using default parameters, hidden here
     *************************************************************/

    /* dump if required by user: */
    if (bCreateDimacs == TRUE) {
      if ( Bmc_DimacsWriteProblem(szDimacsName, k, CnfClauses, 
				  CnfVars, CnfMaxVar) != SUCCESS ) {
	fprintf(nusmv_stdout, 
"\n*************    WARNING    *************\
 \n An error has occured when writing the file \"%s\".\
 \n DIMACS dumping is not allowed.\
 \n*************  END WARNING  *************\n\n", szDimacsName);
      }
    }

    SimResult = Bmc_CallSimSolver(CnfClauses, CnfVars, CnfMaxVar, lsSimResult);

    /*************************************************************
     * PROCESS SOLVER RESULTS
     *************************************************************/
    
    /* Collect all results: */
    if ( (SatoResult==SAT_UNAVAILABLE) || (SatoResult==SimResult) ) {
      result=SimResult; 
    }
    else {
      fprintf( nusmv_stdout, 
	       "Solver: SATO AND SIM RESULTS ARE IN CLASH\n" );
      result = SAT_INTERNAL_ERROR;
    }

    if (result == SAT_INTERNAL_ERROR) {
      fprintf( nusmv_stdout, "Solver: UNRECOVERABLE INTERNAL ERROR\n" );
    }
    else if (result == SAT_UNSATISFIABLE_PROBLEM) {
      if (opt_verbose_level_gt(options, 0)) {
	fprintf( nusmv_stdout, "Solver: given problem is unsatisfiable\n" );
      }
    }
    else {
      /* problem here is satisfiable */
      nusmv_assert(result == SAT_SATISFIABLE_PROBLEM);
      if (opt_verbose_level_gt(options, 0)) {
	fprintf( nusmv_stdout, "Solver: given problem is satisfiable\n" );
      }
      *model = BmcPropModel2SMVModel(lsSimResult, CnfMaxVar, k);
      /* destructs temporary structure: */
      lsDestroy(lsSimResult, NULL);
    }
  }
  
  /* clean up all unnecessary structures */
  BmcIntFreeListOfLists(CnfClauses);
  lsDestroy(CnfVars, NULL);

  return result;
}

/* Example of mapping of STATE VARIABLES and RBC VARIABLES:
    ----------------------------------------------------------
    VAR        x    y    x'   y'   x0   y0   x1   y1   x2   y2
    RBC index  0    1    2    3    4    5    6    7    8    9 
    CNF index  #    1    2    3    4    5    6    7    8    9 
    Time         -2    |   -1   |     0   |     1   |     2   |
    Varindex   0    1  | 0    1 |  0    1 |  0    1 |  0    1 |
*/

static node_ptr BmcPropModel2SMVModel(lsList SimResult, int CnfMaxVar, int k) 
{
  int statevarsnum = Bmc_VarsMgrGetNum();
  lsGen genLit;
  int cnfvar;
  int lit;
  int rbcvar;
  int ** model; /* array of arrays of int */ 
  int time;
  int varindex;
  node_ptr trace = LS_NIL; /* variable for trace, list of bdd states */

  if (opt_verbose_level_gt(options, 1)) {
    fprintf(nusmv_stderr, "The returned CNF model is the following:\n (");
    genLit = lsStart(SimResult);
    cnfvar = 1;
    while (lsNext(genLit, (lsGeneric*) &lit, LS_NH) == LS_OK) {
      fprintf(nusmv_stderr, "%d ", lit);
    }

  lsFinish(genLit);
  fprintf(nusmv_stderr, " )\n");
  }


  if (opt_verbose_level_gt(options, 0)) {
    fprintf( nusmv_stderr, 
	     "**** Converting propositional model into SMV trace\n" );
  }

  if (opt_verbose_level_gt(options, 1)) {
    fprintf(nusmv_stderr, "Number of State Variables: %d, ", statevarsnum);
    fprintf(nusmv_stderr, "time Steps from 0 to %d\n", k);
  }

  /* allocates memory for model, and initialize all locations
     (default value for unassigned literals is false) */
  model = ALLOC(int *, (k + 1));
  nusmv_assert(model != NULL);
  for(time = 0; time <= k; time++) {
    model[time] = ALLOC(int, statevarsnum);
    nusmv_assert(model[time] != NULL);
    /* resets the allocated array: */
    for (varindex = 0; varindex < statevarsnum; varindex++) {
      model[time][varindex] = 0;
    }
  }

  /* scan list of literals */
  genLit = lsStart(SimResult);
  while (lsNext(genLit, (lsGeneric*) &lit, LS_NH) == LS_OK) {
    cnfvar = abs(lit);
    rbcvar = cnfvar;

    if ( (cnfvar >= (2 * statevarsnum)) && 
	 (cnfvar <  ((2 + k + 1) * statevarsnum)) ) {

      /* here only state variables indexes are allowed */
      time = (rbcvar / statevarsnum) - 2;
      varindex = rbcvar % statevarsnum;
      model[time][varindex] = (lit > 0) ? 1 : 0;

      if (opt_verbose_level_gt(options, 1)) {
	fprintf(nusmv_stderr, "Literal %5d: ", lit);
	fprintf( nusmv_stderr, "model[%d][%d] = % d      ", time, varindex,
		 model[time][varindex] );
	print_node(nusmv_stderr, Bmc_GetIthStatevar(varindex));
	fprintf(nusmv_stderr, " == %d", model[time][varindex]);
	fprintf(nusmv_stderr, "\n");
      }
    }
  }

  lsFinish(genLit);

  if (opt_verbose_level_gt(options, 1)) {
	fprintf(nusmv_stderr, "\n");
  }

  /*************************************************************
   * BUILD COUNTEREXAMPLE AS SEQUENCE OF BDDs
   *************************************************************/
  for(time = 0; time <= k; time++) {
    node_ptr state_exp = new_node(TRUEEXP, Nil, Nil);
    
    if (opt_verbose_level_gt(options, 1)) {
      fprintf(nusmv_stderr, " --------------- Time %d --------------- \n", 
	      time);
    }
    
    for(varindex = 0; varindex < (statevarsnum); varindex++) { 
      node_ptr var_exp;
      
      if(model[time][varindex]) {
	var_exp = Bmc_GetIthStatevar(varindex);
      }
      else {
	var_exp = new_node(NOT, Bmc_GetIthStatevar(varindex), Nil);
      }
      state_exp = new_node(AND, var_exp, state_exp);
    }
    
    {
      bdd_ptr state_bdd = (bdd_ptr) eval_spec( Prop_MasterGetBddFsm(), 
					       state_exp, Nil );
      if (opt_verbose_level_gt(options, 1)) {
	print_state(state_bdd, all_symbols, 0);
      }
      trace = cons((node_ptr)state_bdd, trace);
    }
  } /* end of for time */

  /* destroys previously allocated model array: */
  for(time = 0; time <= k; time++) {
    FREE(model[time]);
  }
  FREE(model);

  return(reverse(trace));
}



/**Function********************************************************************

  Synopsis           [BmcSatSolveInvar]

  Description        []

  SideEffects        [pnSimResult will change to result if prob is 
  satisfiable, NULL otherwise]

  SeeAlso            []

******************************************************************************/

static sat_result BmcSatSolveInvar(rbc_ptr prob, int bCreateDimacs, 
				   node_ptr* pnSimResult)
{
  return BmcSatSolveFalsificationProblem(prob, 1, BMC_NO_LOOP, INVARPB_MODE, 
					 bCreateDimacs, pnSimResult);
}


/**Function********************************************************************

  Synopsis           [Builds the RBC model, applying the cone of influence
  if it has been enabled by user.]

  Description        [Applies COI if it is enabled. Uses previuosly built model
  if present.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/

static Fsm_RbcPtr Prop_GetRbcFsmCoi(Prop_Ptr prop) {
  Fsm_RbcPtr fsm;
  node_ptr spec;

  /* checks that a property was selected: */
  nusmv_assert (prop != (Prop_Ptr)NULL);

  fsm = Prop_GetRbcFsm(prop);
  spec = Prop_GetProp(prop);

  if (fsm == (Fsm_RbcPtr)NULL) {

    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Building the Rbc model...\n");
    }
    if (opt_cone_of_influence(options) == true) {
      add_ptr one = add_one(dd_manager);
      node_ptr fc = cmp_struct_get_fair(cmps);
      Set_t spec_dep = Formulae_GetDependencies(spec, fc);
      Set_t cone = ComputeCOI(spec_dep);

      if (opt_verbose_level_gt(options, 0)) {
	fprintf(nusmv_stderr, "Cone of influence\n");
      }

      Prop_BuildFsm(prop, Set_Set2List(cone), fc, one);
      Prop_SetCone(prop, cone);

      { /* constructing the fsm in rbc format */
	Fsm_SexpPtr bfsm = Prop_GetBoolFsm(prop);
	node_ptr init_bexp = Compile_FsmSexpGetInit(bfsm);
	node_ptr trans_bexp = Compile_FsmSexpGetTrans(bfsm);
	node_ptr invar_bexp = Compile_FsmSexpGetInvar(bfsm);
	node_ptr fairness_bexp = Compile_FsmSexpGetFairness(bfsm);
	
	Fsm_RbcPtr rbc_fsm = Compile_FsmRbcAlloc(); /* Create empty RBC FSM */
	
	/* converts fields to rbc and fills the structure: */
	Compile_FsmRbcSetInit( rbc_fsm, Bmc_Bexp2Rbc(init_bexp) );
	Compile_FsmRbcSetTrans( rbc_fsm, Bmc_Bexp2Rbc(trans_bexp) );
	Compile_FsmRbcSetInvar( rbc_fsm, Bmc_Bexp2Rbc(invar_bexp) );
	Compile_FsmRbcSetFairness( rbc_fsm, Bmc_MapBexp2Rbc(fairness_bexp) );

	Prop_SetRbcFsm(prop, rbc_fsm);
      }

      fsm = Prop_GetRbcFsm(prop);
    }
    else {
      if (opt_verbose_level_gt(options, 0)) {
	fprintf(nusmv_stderr, "No cone of influence, using whole model\n");
      }
      Prop_SetFsmToMaster(prop);
      fsm = Prop_GetRbcFsm(prop);
    }
  }
  else {
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Using previously built Rbc model...\n");
    }
  }
  return(fsm);
}



static void 
BmcExpandMacrosInFilename(const char* filename, char* filename_expanded, 
			  size_t buf_len)
{
  char szBuffer[1024];
  char szLoopback[16]; 

  static SubstString aSubstTable[] =  
    {
      SYMBOL_CREATE("@F"),
      SYMBOL_CREATE("@f"),
      SYMBOL_CREATE("@k"),
      SYMBOL_CREATE("@l"),
      SYMBOL_CREATE("@n"), 
      SYMBOL_CREATE("@@")
    };
 

  /* customizes the table with runtime values: */
  Bmc_StripPathExt(get_input_file(options), szBuffer);
  Bmc_ConvertLoopbackFromInteger(get_bmc_pb_loop(options), szLoopback, 
				 sizeof(szLoopback));

  SYMBOL_ASSIGN(aSubstTable[0], string,  "%s", get_input_file(options));
  SYMBOL_ASSIGN(aSubstTable[1], string,  "%s", szBuffer);
  SYMBOL_ASSIGN(aSubstTable[2], integer, "%d", get_bmc_pb_length(options));
  SYMBOL_ASSIGN(aSubstTable[3], string, "%s", szLoopback);
  if (bmcGetCurrentPropertyIndex() != BMC_NO_PROPERTY) {
    SYMBOL_ASSIGN(aSubstTable[4], integer, "%d", bmcGetCurrentPropertyIndex());
  }
  else {
    SYMBOL_ASSIGN(aSubstTable[4], string, "%s", "undef");
  }
  SYMBOL_ASSIGN(aSubstTable[5], string,  "%s", "@");

  {
    int i;
    /* copy the source string into the destination one: */
    strncpy(filename_expanded, filename, buf_len);

    for(i=0; i < (sizeof(aSubstTable)/sizeof(SubstString)); ++i) {
      apply_string_macro_expansion( &(aSubstTable[i]), 
				    filename_expanded,
				    buf_len );
    } /* for each symbol template */
  }
}


/**Function********************************************************************

  Synopsis           [Checks the given LTL specification, or all LTL 
  specifications if no formula is given. Parameters are the maximum 
  length and the loopback values]

  Description        [This function saves the current values of environment 
  variables <i>bmc_length</i>, <i>bmc_loopback</i> and 
  <i>bmc_dimacs_filename</i>. Then after command line processing it calls
  the function Bmc_GenSolvePbs which generates and solve all problems from 
  0 to k. <BR>
  Before exit this function restores old values of all modified environment
  variables.]

  SideEffects        [Property database may change]

  SeeAlso            [CommandCheckLtlSpecBmcOnePb, Bmc_GenSolvePbs]

  CommandName        [check_ltlspec_bmc] 	   

  CommandSynopsis    [Checks the given LTL specification, or all LTL 
  specifications if no formula is given. Checking parameters are the maximum 
  length and the loopback values]  

  CommandArguments   [\[-h | -n idx | -p "formula" \[IN context\]\] \[-k max_length\] \[-l loopback\] \[-o filename\]]  

  CommandDescription [
  This command generates one or more problems, and calls 
  SAT solver for each one. Each problem is related to a specific problem 
  length, which increases from zero (0) to the given maximum problem 
  length. Here "<i>length</i>" is the length of the problem that system 
  is going to generate and/or solve. <BR>
  In this context the maximum problem length is represented by the 
  <i>-k</i> command parameter, or by its default value stored in the 
  environment variable <i>bmc_length</i>.<BR>
  The single generated problem also depends on the "<i>loopback</i>"
  parameter you can explicitly specify by the <i>-l</i> option, or by its
  default value stored in the environment variable <i>bmc_loopback</i>. <BR>
  The property to be checked may be specified using the <i>-n idx</i> or  
  the <i>-p "formula"</i> options. <BR>
  If you need to generate a dimacs dump file of all generated problems, you 
  must use the option <i>-o "filename"</i>. <BR>  
  <p>
  Command options:<p>
  <dl>
    <dt> <tt>-n <i>index</i></tt>
       <dd> <i>index</i> is the numeric index of a valid LTL specification 
       formula actually located in the properties database. <BR>
    <dt> <tt>-p "formula" \[IN context\]</tt>
       <dd> Checks the <tt>formula</tt> specified on the command-line. <BR>
            <tt>context</tt> is the module instance name which the variables
            in <tt>formula</tt> must be evaluated in.
    <dt> <tt>-k <i>max_length</i></tt>
       <dd> <i>max_length</i> is the maximum problem length used . Only 
       natural number are valid values for this option. If no value is given
       the environment 
variable <i>bmc_length</i> is considered instead. 
    <dt> <tt>-l <i>loopback</i></tt>
       <dd> <i>loopback</i> value may be: <BR>
       - a natural number in (0, <i>max_length</i>). Positive sign ('+') can 
       be also used as prefix of the number. Invalid combination of length and
       loopback will be skipped during the generation/solving process.<BR>
       - a negative number in (-1, -<i>bmc_length</i>). In this case 
       <i>loopback</i> is considered a value relative to <i>max_length</i>. 
       Invalid combination of length and loopback will be skipped during the 
       generation/solving process.<BR>
       - the symbol 'X', which means "no loopback" <BR>
       - the symbol '*', which means "all possible loopback from zero to 
       <i>length</i>"
    <dt> <tt>-o <i>filename</i></tt>
       <dd> <i>filename</i> is the name of the dumped dimacs file. 
       It may contain special symbols which will be macro-expanded to form 
       the real file name. Possible symbols are: <BR>
       - @F: model name with path part <BR>
       - @f: model name without path part <BR>
       - @k: current problem length <BR>
       - @l: current loopback value <BR>
       - @n: index of the currently processed formula in the properties 
       database <BR>
       - @@: the '@' character
  </dl>
  ]  

******************************************************************************/
static int CommandCheckLtlSpecBmc(int argc, char ** argv)
{
  Prop_Ptr ltlprop=(Prop_Ptr)NULL;   /* The property being processed */
  BMC_CMD_DECL_SAVING_OPTIONS();
  outcome opt_handling_res; 
  int bCreateDimacs = FALSE; /* create dimacs only if -o opt is specified */ 

  BMC_CMD_SAVE_OPTIONS();

  /* ----------------------------------------------------------------------- */
  /* Options handling: */
  opt_handling_res = bmc_CommandsWith_k_l_OptionsHandling(argc, argv, &ltlprop,
							  Prop_Ltl, 
							  &bCreateDimacs);
  if (opt_handling_res == SUCCESS_REQUIRED_HELP) {
    BMC_CMD_RESTORE_OPTIONS(); 
    return UsageBmcCheckLtlSpec();
  }
  if (opt_handling_res != SUCCESS) {
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }
  /* ----------------------------------------------------------------------- */


  /* make sure the model hierarchy has been flattened */
  if (cmp_struct_get_flatten_hrc (cmps) == 0) {
    fprintf (nusmv_stderr, "The hierarchy must be flattened before. Use "\
	     "the \"flatten_hierarchy\" command.\n");
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }

  /* make sure bmc has been set up */
  if (cmp_struct_get_bmc_setup(cmps) == 0) {
    fprintf (nusmv_stderr, "Bmc must be setup before. Use "\
	     "the \"bmc_setup\" command.\n");
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }

  /* prepare the list of properties if no property was selected: */
  if (ltlprop == NIL_PTR(Prop_Ptr)) {
    lsList props = Prop_Db_GetPropsOfType(Prop_Ltl);
    lsGen  iterator; 
    Prop_Ptr prop;

    nusmv_assert(props != LS_NIL);

    lsForEachItem(props, iterator, prop) {
      Bmc_GenSolvePbs((Prop_Ptr)prop, TRUE, bCreateDimacs, TRUE);
    }

    lsDestroy(props, NULL); /* the list is no longer needed */
  }
  else {
    /* its time to solve: */
    Bmc_GenSolvePbs(ltlprop, TRUE, bCreateDimacs, TRUE);
  }

  BMC_CMD_RESTORE_OPTIONS(); 
  return 0; 
}


/**Function********************************************************************

  Synopsis           [Usage string for command check_ltlspec_bmc]

  Description        []

  SideEffects        [None]

  SeeAlso            [CommandCheckLtlSpecBmc]

******************************************************************************/
static int UsageBmcCheckLtlSpec(void)
{
  fprintf(nusmv_stderr, "\nUsage: check_ltlspec_bmc [-h | -n idx | -p \"formula\"] [-k max_length] [-l loopback]\n\t\t\t [-o <filename>]\n");
  fprintf(nusmv_stderr, "  -h \t\tPrints the command usage\n");
  fprintf(nusmv_stderr, "  -n idx\tChecks the LTL property specified with <idx>\n");
  fprintf(nusmv_stderr, "  -p \"formula\"\tChecks the specified LTL property\n");
  fprintf(nusmv_stderr, "\t\tIf no property is specified, checks all LTL properties.\n"); 
  fprintf(nusmv_stderr, "  -k max_length\tChecks the property using <max_length> value instead of using the\n\t\tvariable <bmc_length> value\n");
  fprintf(nusmv_stderr, "  -l loopback\tChecks the property using <loopback> value instead of using the\n\t\tvariable <bmc_loopback> value\n");
  fprintf(nusmv_stderr, "  -o filename\tGenerates dimacs output file too. <filename> may contain patterns\n\n"); 

  return(1);
}



/**Function********************************************************************

  Synopsis           [Checks the given LTL specification, or all LTL 
  specifications if no formula is given. Checking parameters are the problem 
  length and the loopback values]

  Description        [This function saves the current values of environment 
  variables <i>bmc_length</i>, <i>bmc_loopback</i> and 
  <i>bmc_dimacs_filename</i>. Then after command line processing it calls
  the function Bmc_GenSolvePbs which generates and solve the singleton 
  problem of length k and loopback l. <BR>
  Before exit this function restores old values of all modified environment
  variables.]

  SideEffects        [Property database may change]

  SeeAlso            [CommandCheckLtlSpecBmc, Bmc_GenSolvePbs]

  CommandName        [check_ltlspec_bmc_onepb] 	   

  CommandSynopsis    [Checks the given LTL specification, or all LTL 
  specifications if no formula is given. Checking parameters are the single
  problem length and the loopback values]  

  CommandArguments   [\[-h | -n idx | -p "formula" \[IN context\]\] \[-k length\] \[-l loopback\] \[-o filename\]]  

  CommandDescription [As command check_ltlspec_bmc but it produces only one 
  single problem with fixed length and loopback values, with no iteration
  of the problem length from zero to max_length. <BR>  
  <p>
  Command options:<p>
  <dl>
    <dt> <tt>-n <i>index</i></tt>
       <dd> <i>index</i> is the numeric index of a valid LTL specification 
       formula actually located in the properties database. <BR>
       The validity of <i>index</i> value is checked out by the system.
    <dt> <tt>-p "formula \[IN context\]"</tt>
       <dd> Checks the <tt>formula</tt> specified on the command-line. <BR>
            <tt>context</tt> is the module instance name which the variables
            in <tt>formula</tt> must be evaluated in.
    <dt> <tt>-k <i>length</i></tt>
       <dd> <i>length</i> is the problem length used when generating the 
       single problem. Only natural number are valid values for this option. 
       If no value is given the environment variable <i>bmc_length</i> is 
       considered instead. 
    <dt> <tt>-l <i>loopback</i></tt>
       <dd> <i>loopback</i> value may be: <BR>
       - a natural number in (0, <i>max_length</i>). Positive sign ('+') can 
       be also used as prefix of the number. Invalid combination of length and
       loopback will be skipped during the generation/solving process.<BR>
       - a negative number in (-1, -<i>bmc_length</i>). In this case 
       <i>loopback</i> is considered a value relative to <i>length</i>. 
       Invalid combination of length and loopback will be skipped during the 
       generation/solving process.<BR>
       - the symbol 'X', which means "no loopback" <BR>
       - the symbol '*', which means "all possible loopback from zero to 
       <i>length</i>"
    <dt> <tt>-o <i>filename</i></tt>
       <dd> <i>filename</i> is the name of the dumped dimacs file. 
       It may contain special symbols which will be macro-expanded to form 
       the real file name. Possible symbols are: <BR>
       - @F: model name with path part <BR>
       - @f: model name without path part <BR>
       - @k: current problem length <BR>
       - @l: current loopback value <BR>
       - @n: index of the currently processed formula in the properties 
       database <BR>
       - @@: the '@' character
  </dl>]  

******************************************************************************/
static int CommandCheckLtlSpecBmcOnePb(int argc, char ** argv)
{
  Prop_Ptr ltlprop=(Prop_Ptr)NULL;   /* The property being processed */
  BMC_CMD_DECL_SAVING_OPTIONS();
  outcome opt_handling_res; 
  int bCreateDimacs = FALSE; /* create dimacs only if -o opt is specified */ 

  BMC_CMD_SAVE_OPTIONS();

  /* ----------------------------------------------------------------------- */
  /* Options handling: */
  opt_handling_res = bmc_CommandsWith_k_l_OptionsHandling(argc, argv, &ltlprop,
							  Prop_Ltl, 
							  &bCreateDimacs);
  if (opt_handling_res == SUCCESS_REQUIRED_HELP) {
    BMC_CMD_RESTORE_OPTIONS(); 
    return UsageBmcCheckLtlSpecOnePb();
  }
  if (opt_handling_res != SUCCESS) {
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }
  /* ----------------------------------------------------------------------- */


  /* make sure the model hierarchy has been flattened */
  if (cmp_struct_get_flatten_hrc (cmps) == 0) {
    fprintf (nusmv_stderr, "The hierarchy must be flattened before. Use "\
	     "the \"flatten_hierarchy\" command.\n");
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }

  /* make sure bmc has been set up */
  if (cmp_struct_get_bmc_setup(cmps) == 0) {
    fprintf (nusmv_stderr, "Bmc must be setup before. Use "\
	     "the \"bmc_setup\" command.\n");
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }

  /* prepare the list of properties if no property was selected: */
  if (ltlprop == NIL_PTR(Prop_Ptr)) {
    lsList props = Prop_Db_GetPropsOfType(Prop_Ltl);
    lsGen  iterator; 
    Prop_Ptr prop;

    nusmv_assert(props != LS_NIL);

    lsForEachItem(props, iterator, prop) {
      Bmc_GenSolvePbs((Prop_Ptr)prop, FALSE, bCreateDimacs, TRUE);
    }

    lsDestroy(props, NULL); /* the list is no longer needed */
  }
  else {
    /* its time to solve: */
    Bmc_GenSolvePbs(ltlprop, FALSE, bCreateDimacs, TRUE);
  }

  BMC_CMD_RESTORE_OPTIONS(); 
  return 0; 
}


/**Function********************************************************************

  Synopsis           [Usage string for command check_ltlspec_bmc_onepb]

  Description        []

  SideEffects        [None]

  SeeAlso            [CommandCheckLtlSpecBmcOnePb]

******************************************************************************/
static int UsageBmcCheckLtlSpecOnePb(void)
{
  fprintf(nusmv_stderr, "\nUsage: check_ltlspec_bmc_onepb [-h | -n idx | -p \"formula\"] [-k length] [-l loopback]\n\t\t\t [-o filename]\n");
  fprintf(nusmv_stderr, "  -h \t\tPrints the command usage\n");
  fprintf(nusmv_stderr, "  -n idx\tChecks the LTL property specified with <idx>\n");
  fprintf(nusmv_stderr, "  -p \"formula\"\tChecks the specified LTL property\n");
  fprintf(nusmv_stderr, "\t\tIf no property is specified, checks all LTL properties.\n"); 
  fprintf(nusmv_stderr, "  -k length\tChecks the property using <length> value instead of using the\n\t\tvariable <bmc_length> value\n");
  fprintf(nusmv_stderr, "  -l loopback\tChecks the property using <loopback> value instead of using the\n\t\tvariable <bmc_loopback> value\n");
  fprintf(nusmv_stderr, "  -o filename\tGenerates dimacs output file too. <filename> may contain patterns\n\n"); 

  return(1);
}


/**Function********************************************************************

  Synopsis           [Generates only one problem of fixed length and 
  loopback, and dumps the problem to a dimacs file. The single problem 
  is dumped for the given LTL specification, or for all LTL 
  specifications if no formula is given]

  Description        [This function saves the current values of environment 
  variables <i>bmc_length</i>, <i>bmc_loopback</i> and 
  <i>bmc_dimacs_filename</i>. Then after command line processing it calls
  the function Bmc_GenSolvePbs to generate and dump the single problem.<BR>
  Before exit this function restores old values of all modified environment
  variables.]

  SideEffects        [Property database may change]

  SeeAlso            [CommandGenLtlSpecBmc, Bmc_GenSolvePbs]

  CommandName        [gen_ltlspec_bmc_onepb] 	   

  CommandSynopsis    [Dumps into one dimacs file the problem generated for
  the given LTL specification, or for all LTL specifications if no formula 
  is explicitly given. 
  Generation and dumping parameters are the problem length and the loopback 
  values]  

  CommandArguments   [\[-h | -n idx | -p "formula" \[IN context\]\] \[-k length\] 
  \[-l loopback\] \[-o filename\]]  

  CommandDescription [ As the <i>gen_ltlspec_bmc</i> command, but it generates
  and dumps only one problem given its length and loopback. <BR>
  <p>
  Command options:<p>
  <dl>
    <dt> <tt>-n <i>index</i></tt>
       <dd> <i>index</i> is the numeric index of a valid LTL specification 
       formula actually located in the properties database. <BR>
       The validity of <i>index</i> value is checked out by the system.
    <dt> <tt>-p "formula \[IN context\]"</tt>
       <dd> Checks the <tt>formula</tt> specified on the command-line. <BR>
            <tt>context</tt> is the module instance name which the variables
            in <tt>formula</tt> must be evaluated in.
    <dt> <tt>-k <i>length</i></tt>
       <dd> <i>length</i> is the single problem length used to generate and 
       dump it. Only natural number are valid values for this option. 
       If no value is given the environment variable <i>bmc_length</i> 
       is considered instead. 
    <dt> <tt>-l <i>loopback</i></tt>
       <dd> <i>loopback</i> value may be: <BR>
       - a natural number in (0, <i>length</i>). Positive sign ('+') can 
       be also used as prefix of the number. Invalid combination of length and
       loopback will be skipped during the generation and dumping process.<BR>
       - a negative number in (-1, -<i>length</i>). 
       Invalid combination of length and loopback will be skipped during 
       the generation process.<BR>
       - the symbol 'X', which means "no loopback" <BR>
       - the symbol '*', which means "all possible loopback from zero to 
       <i>length</i>"
    <dt> <tt>-o <i>filename</i></tt>
       <dd> <i>filename</i> is the name of the dumped dimacs file. If this
       options is not specified, variable <i>bmc_dimacs_filename</i> will be
       considered. The file name string may contain special symbols which 
       will be macro-expanded to form the real file name. 
       Possible symbols are: <BR>
       - @F: model name with path part <BR>
       - @f: model name without path part <BR>
       - @k: current problem length <BR>
       - @l: current loopback value <BR>
       - @n: index of the currently processed formula in the properties 
       database <BR>
       - @@: the '@' character
  </dl>]  

******************************************************************************/
static int CommandGenLtlSpecBmcOnePb(int argc, char ** argv)
{
  Prop_Ptr ltlprop=(Prop_Ptr)NULL;   /* The property being processed */
  BMC_CMD_DECL_SAVING_OPTIONS();
  outcome opt_handling_res; 
  int bCreateDimacs = TRUE; /* create dimacs in any case */

  BMC_CMD_SAVE_OPTIONS();

  /* ----------------------------------------------------------------------- */
  /* Options handling: */
  opt_handling_res = bmc_CommandsWith_k_l_OptionsHandling(argc, argv, &ltlprop,
							  Prop_Ltl, 
							  &bCreateDimacs);
  if (opt_handling_res == SUCCESS_REQUIRED_HELP) {
    BMC_CMD_RESTORE_OPTIONS(); 
    return UsageBmcGenLtlSpecOnePb();
  }
  if (opt_handling_res != SUCCESS) {
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }
  /* ----------------------------------------------------------------------- */


  /* make sure the model hierarchy has been flattened */
  if (cmp_struct_get_flatten_hrc (cmps) == 0) {
    fprintf (nusmv_stderr, "The hierarchy must be flattened before. Use "\
	     "the \"flatten_hierarchy\" command.\n");
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }

  /* make sure bmc has been set up */
  if (cmp_struct_get_bmc_setup(cmps) == 0) {
    fprintf (nusmv_stderr, "Bmc must be setup before. Use "\
	     "the \"bmc_setup\" command.\n");
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }

  /* prepare the list of properties if no property was selected: */
  if (ltlprop == NIL_PTR(Prop_Ptr)) {
    lsList props = Prop_Db_GetPropsOfType(Prop_Ltl);
    lsGen  iterator; 
    Prop_Ptr prop;

    nusmv_assert(props != LS_NIL);

    lsForEachItem(props, iterator, prop) {
      Bmc_GenSolvePbs((Prop_Ptr)prop, FALSE, bCreateDimacs, FALSE);
    }

    lsDestroy(props, NULL); /* the list is no longer needed */
  }
  else {
    /* its time to solve: */
    Bmc_GenSolvePbs(ltlprop, FALSE, bCreateDimacs, FALSE);
  }

  BMC_CMD_RESTORE_OPTIONS(); 
  return 0; 
}


/**Function********************************************************************

  Synopsis           [Usage string for command gen_ltlspec_bmc_onepb]

  Description        []

  SideEffects        [None]

  SeeAlso            [CommandGenLtlSpecBmcOnePb]

******************************************************************************/
static int UsageBmcGenLtlSpecOnePb(void)
{
  fprintf(nusmv_stderr, "\nUsage: gen_ltlspec_bmc_onepb [-h | -n idx | -p \"formula\"] [-k length] [-l loopback]\n\t\t\t [-o filename]\n");
  fprintf(nusmv_stderr, "  -h \t\tPrints the command usage\n");
  fprintf(nusmv_stderr, "  -n idx\tChecks the LTL property specified with <idx>\n");
  fprintf(nusmv_stderr, "  -p \"formula\"\tChecks the specified LTL property\n");
  fprintf(nusmv_stderr, "\t\tIf no property is specified, checks all LTL properties.\n"); 
  fprintf(nusmv_stderr, "  -k length\tChecks the property using <length> value instead of using the\n\t\tvariable <bmc_length> value\n");
  fprintf(nusmv_stderr, "  -l loopback\tChecks the property using <loopback> value instead of using the\n\t\tvariable <bmc_loopback> value\n");
  fprintf(nusmv_stderr, "  -o filename\tUses <filename> as dimacs file instead of \"bmc_dimacs_filename\"\t\t variable. <filename> may contain patterns\n\n"); 

  return(1);
}


/**Function********************************************************************

  Synopsis           [Generates length_max+1 problems iterating length from 
  zero to length_max, and dumps each problem to a dimacs file. Each problem 
  is dumped for the given LTL specification, or for all LTL 
  specifications if no formula is given. Generation parameters are the maximum 
  length and the loopback values]

  Description        [This function saves the current values of environment 
  variables <i>bmc_length</i>, <i>bmc_loopback</i> and 
  <i>bmc_dimacs_filename</i>. Then after command line processing it calls
  the function Bmc_GenSolvePbs to generate and dump all problems from 
  0 to k. <BR>
  Before exit this function restores old values of all modified environment
  variables.]

  SideEffects        [Property database may change]

  SeeAlso            [CommandGenLtlSpecBmcOnePb, Bmc_GenSolvePbs]

  CommandName        [gen_ltlspec_bmc] 	   

  CommandSynopsis    [Dumps into one or more dimacs files the given LTL 
  specification, or all LTL specifications if no formula is given. 
  Generation and dumping parameters are the maximum length and the loopback 
  values]  

  CommandArguments   [\[-h | -n idx | -p "formula" \[IN context\]\] \[-k max_length\] 
  \[-l loopback\] \[-o filename\]]  

  CommandDescription [  This command generates one or more problems, and  
  dumps each problem into a dimacs file. Each problem is related to a specific
  problem length, which increases from zero (0) to the given maximum problem 
  length. In this short description "<i>length</i>" is the length of the 
  problem that system is going to dump out. <BR>
  In this context the maximum problem length is represented by the 
  <i>max_length</i> parameter, or by its default value stored in the 
  environment variable <i>bmc_length</i>.<BR>
  Each dumped problem also depends on the loopback you can explicitly 
  specify by the <i>-l</i> option, or by its default value stored in the 
  environment variable <i>bmc_loopback</i>. <BR>
  The property to be checked may be specified using the <i>-n idx</i> or  
  the <i>-p "formula"</i> options. <BR>
  You may specify dimacs file name by using the option <i>-o "filename"</i>, 
  otherwise the default value stored in the environment variable 
  <i>bmc_dimacs_filename</i> will be considered.<BR>  
  <p>
  Command options:<p>
  <dl>
    <dt> <tt>-n <i>index</i></tt>
       <dd> <i>index</i> is the numeric index of a valid LTL specification 
       formula actually located in the properties database. <BR>
       The validity of <i>index</i> value is checked out by the system.
    <dt> <tt>-p "formula \[IN context\]"</tt>
       <dd> Checks the <tt>formula</tt> specified on the command-line. <BR>
            <tt>context</tt> is the module instance name which the variables
            in <tt>formula</tt> must be evaluated in.
    <dt> <tt>-k <i>max_length</i></tt>
       <dd> <i>max_length</i> is the maximum problem length used when 
       increasing problem length starting from zero. Only natural number are
       valid values for this option. If no value is given the environment 
       variable <i>bmc_length</i> value is considered instead. 
    <dt> <tt>-l <i>loopback</i></tt>
       <dd> <i>loopback</i> value may be: <BR>
       - a natural number in (0, <i>max_length</i>). Positive sign ('+') can 
       be also used as prefix of the number. Invalid combination of length and
       loopback will be skipped during the generation and dumping process.<BR>
       - a negative number in (-1, -<i>bmc_length</i>). In this case 
       <i>loopback</i> is considered a value relative to <i>max_length</i>. 
       Invalid combination of length and loopback will be skipped during 
       the generation process.<BR>
       - the symbol 'X', which means "no loopback" <BR>
       - the symbol '*', which means "all possible loopback from zero to 
       <i>length</i>"
    <dt> <tt>-o <i>filename</i></tt>
       <dd> <i>filename</i> is the name of dumped dimacs files. If this
       options is not specified, variable <i>bmc_dimacs_filename</i> will be
       considered. The file name string may contain special symbols which 
       will be macro-expanded to form the real file name. 
       Possible symbols are: <BR>
       - @F: model name with path part <BR>
       - @f: model name without path part <BR>
       - @k: current problem length <BR>
       - @l: current loopback value <BR>
       - @n: index of the currently processed formula in the properties 
       database <BR>
       - @@: the '@' character
  </dl>]  
   

******************************************************************************/
static int CommandGenLtlSpecBmc(int argc, char ** argv)
{
  Prop_Ptr ltlprop=(Prop_Ptr)NULL;   /* The property being processed */
  BMC_CMD_DECL_SAVING_OPTIONS();
  outcome opt_handling_res; 
  int bCreateDimacs = TRUE; /* create dimacs in any case */

  BMC_CMD_SAVE_OPTIONS();

  /* ----------------------------------------------------------------------- */
  /* Options handling: */
  opt_handling_res = bmc_CommandsWith_k_l_OptionsHandling(argc, argv, &ltlprop,
							  Prop_Ltl, 
							  &bCreateDimacs);
  if (opt_handling_res == SUCCESS_REQUIRED_HELP) {
    BMC_CMD_RESTORE_OPTIONS(); 
    return UsageBmcGenLtlSpec();
  }
  if (opt_handling_res != SUCCESS) {
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }
  /* ----------------------------------------------------------------------- */


  /* make sure the model hierarchy has been flattened */
  if (cmp_struct_get_flatten_hrc (cmps) == 0) {
    fprintf (nusmv_stderr, "The hierarchy must be flattened before. Use "\
	     "the \"flatten_hierarchy\" command.\n");
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }

  /* make sure bmc has been set up */
  if (cmp_struct_get_bmc_setup(cmps) == 0) {
    fprintf (nusmv_stderr, "Bmc must be setup before. Use "\
	     "the \"bmc_setup\" command.\n");
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }

  /* prepare the list of properties if no property was selected: */
  if (ltlprop == NIL_PTR(Prop_Ptr)) {
    lsList props = Prop_Db_GetPropsOfType(Prop_Ltl);
    lsGen  iterator; 
    Prop_Ptr prop;

    nusmv_assert(props != LS_NIL);

    lsForEachItem(props, iterator, prop) {
      Bmc_GenSolvePbs((Prop_Ptr)prop, TRUE, bCreateDimacs, FALSE);
    }

    lsDestroy(props, NULL); /* the list is no longer needed */
  }
  else {
    /* its time to solve: */
    Bmc_GenSolvePbs(ltlprop, TRUE, bCreateDimacs, FALSE);
  }

  BMC_CMD_RESTORE_OPTIONS(); 
  return 0; 
}

/**Function********************************************************************

  Synopsis           [Usage string for command gen_ltlspec_bmc]

  Description        []

  SideEffects        [None]

  SeeAlso            [CommandGenLtlSpecBmc]

******************************************************************************/
static int UsageBmcGenLtlSpec(void)
{
  fprintf(nusmv_stderr, "\nUsage: gen_ltlspec_bmc [-h | -n idx | -p \"formula\"] [-k max_length] [-l loopback]\n\t\t\t [-o filename]\n");
  fprintf(nusmv_stderr, "  -h \t\tPrints the command usage\n");
  fprintf(nusmv_stderr, "  -n idx\tChecks the LTL property specified with <idx>\n");
  fprintf(nusmv_stderr, "  -p \"formula\"\tChecks the specified LTL property\n");
  fprintf(nusmv_stderr, "\t\tIf no property is specified, checks all LTL properties.\n"); 
  fprintf(nusmv_stderr, "  -k max_length\tChecks the property using <max_length> value instead of using the\n\t\tvariable <bmc_length> value\n");
  fprintf(nusmv_stderr, "  -l loopback\tChecks the property using <loopback> value instead of using the\n\t\tvariable <bmc_loopback> value\n");
  fprintf(nusmv_stderr, "  -o filename\tUses <filename> as dimacs file instead of using the\n\t\t\"bmc_dimacs_filename\" variable. <filename> may contain patterns\n\n"); 

  return(1);
}


/**Function********************************************************************

  Synopsis           [Generates and dumps the problem for the given 
  invariant or for all invariants if no formula is given. SAT solver is not
  invoked.]

  Description        [After command line processing calls Bmc_GenSolveInvar 
  to dump the generated invariant problem. 
  If you specify the <i>-o "filename"</i> option a dimacs file named  
  "filename" will be created, otherwise the environment variable 
  <i>bmc_invar_dimacs_filename</i> value will be considered.]

  SideEffects        [Property database may change]

  SeeAlso            [Bmc_GenSolveInvar]

  CommandName        [gen_invar_bmc] 	   

  CommandSynopsis    [Generates the given invariant, or all 
  invariants if no formula is given]  

  CommandArguments   [\[-h | -n idx | -p "formula" \[IN context\]\] \[-o filename\]]  

  CommandDescription [<p>
  Command options:<p>
  <dl>
    <dt> <tt>-n <i>index</i></tt>
       <dd> <i>index</i> is the numeric index of a valid INVAR specification 
       formula actually located in the properties database. <BR>
       The validity of <i>index</i> value is checked out by the system.
    <dt> <tt>-p "formula" \[IN context\]</tt>
       <dd> Checks the <tt>formula</tt> specified on the command-line. <BR>
            <tt>context</tt> is the module instance name which the variables
            in <tt>formula</tt> must be evaluated in.
    <dt> <tt>-o <i>filename</i></tt>
       <dd> <i>filename</i> is the name of the dumped dimacs file. If you 
       do not use this option the dimacs file name is taken from the 
       environment variable <i>bmc_invar_dimacs_filename</i>. <BR> 
       File name may contain special symbols which will be macro-expanded 
       to form the real dimacs file name. Possible symbols are: <BR>
       - @F: model name with path part <BR>
       - @f: model name without path part <BR>
       - @n: index of the currently processed formula in the properties 
       database <BR>
       - @@: the '@' character
  </dl>]  

******************************************************************************/
static int CommandGenInvarBmc(int argc, char ** argv)
{
  Prop_Ptr invarprop=(Prop_Ptr)NULL;   /* The property being processed */
  BMC_CMD_DECL_SAVING_OPTIONS();
  outcome opt_handling_res; 
  int bCreateDimacs = TRUE; /* create dimacs in any case */

  BMC_CMD_SAVE_OPTIONS();

  /* ----------------------------------------------------------------------- */
  /* Options handling: */
  opt_handling_res = 
   bmc_CommandsWithout_k_l_OptionsHandling(argc, argv, &invarprop, Prop_Invar,
					   &bCreateDimacs);

  if (opt_handling_res == SUCCESS_REQUIRED_HELP) {
    BMC_CMD_RESTORE_OPTIONS(); 
    return UsageBmcGenInvar();
  }
  if (opt_handling_res != SUCCESS) {
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }
  /* ----------------------------------------------------------------------- */


  /* make sure the model hierarchy has been flattened */
  if (cmp_struct_get_flatten_hrc (cmps) == 0) {
    fprintf (nusmv_stderr, "The hierarchy must be flattened before. Use "\
	     "the \"flatten_hierarchy\" command.\n");
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }

  /* make sure bmc has been set up */
  if (cmp_struct_get_bmc_setup(cmps) == 0) {
    fprintf (nusmv_stderr, "Bmc must be setup before. Use "\
	     "the \"bmc_setup\" command.\n");
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }

  /* prepare the list of properties if no property was selected: */
  if (invarprop == NIL_PTR(Prop_Ptr)) {
    lsList props = Prop_Db_GetPropsOfType(Prop_Invar);
    lsGen  iterator; 
    Prop_Ptr prop;

    nusmv_assert(props != LS_NIL);

    lsForEachItem(props, iterator, prop) {
      Bmc_GenSolveInvar((Prop_Ptr)prop, bCreateDimacs, FALSE);
    }

    lsDestroy(props, NULL); /* the list is no longer needed */
  }
  else {
    /* its time to generate dimacs: */
    Bmc_GenSolveInvar(invarprop, bCreateDimacs, FALSE);
  }

  BMC_CMD_RESTORE_OPTIONS(); 
  return 0; 
}


/**Function********************************************************************

  Synopsis           [Usage string for command gen_invar_bmc]

  Description        []

  SideEffects        [None]

  SeeAlso            [CommandGenInvarBmc]

******************************************************************************/
static int UsageBmcGenInvar(void)
{
  fprintf(nusmv_stderr, "\nUsage: gen_invar_bmc [-h | -n idx | -p \"formula\"] [-o filename]\n");
  fprintf(nusmv_stderr, "  -h \t\tPrints the command usage\n");
  fprintf(nusmv_stderr, "  -n idx\tChecks the INVAR property specified with <idx>\n");
  fprintf(nusmv_stderr, "  -p \"formula\"\tChecks the specified INVAR propositional property\n");
  fprintf(nusmv_stderr, "\t\tIf no property is specified, checks all INVAR properties.\n"); 
  fprintf(nusmv_stderr, "  -o filename\tUses <filename> as dimacs file instead of using the\n\t\t\"bmc_invar_dimacs_filename\" variable. <filename> may contain patterns\n\n"); 

  return(1);
}


/**Function********************************************************************

  Synopsis           [Generates and solve the given invariant, or all 
  invariants if no formula is given]

  Description        [After command line processing calls Bmc_GenSolveInvar 
  to solve and eventually dump the generated invariant problem. If you specify
  the <i>-o "filename"</i> option a dimacs file will be generated, otherwise
  no dimacs dump will be performed]

  SideEffects        [Property database may change]

  SeeAlso            [Bmc_GenSolveInvar]

  CommandName        [check_invar_bmc] 	   

  CommandSynopsis    [Generates and solve the given invariant, or all 
  invariants if no formula is given]  

  CommandArguments   [\[-h | -n idx | -p "formula" \[IN context\]\] \[-o filename\]]  

  CommandDescription [<p>
  Command options:<p>
  <dl>
    <dt> <tt>-n <i>index</i></tt>
       <dd> <i>index</i> is the numeric index of a valid INVAR specification 
       formula actually located in the properties database. <BR>
       The validity of <i>index</i> value is checked out by the system.
    <dt> <tt>-p "formula \[IN context\]"</tt>
       <dd> Checks the <tt>formula</tt> specified on the command-line. <BR>
            <tt>context</tt> is the module instance name which the variables
            in <tt>formula</tt> must be evaluated in.
    <dt> <tt>-o <i>filename</i></tt>
       <dd> <i>filename</i> is the name of the dumped dimacs file. 
       It may contain special symbols which will be macro-expanded to form 
       the real file name. Possible symbols are: <BR>
       - @F: model name with path part <BR>
       - @f: model name without path part <BR>
       - @n: index of the currently processed formula in the properties 
       database <BR>
       - @@: the '@' character
  </dl>]  

******************************************************************************/
static int CommandCheckInvarBmc(int argc, char ** argv)
{
  Prop_Ptr invarprop=(Prop_Ptr)NULL;   /* The property being processed */
  BMC_CMD_DECL_SAVING_OPTIONS();
  outcome opt_handling_res; 
  int bCreateDimacs = FALSE; /* create dimacs in any case */

  BMC_CMD_SAVE_OPTIONS();

  /* ----------------------------------------------------------------------- */
  /* Options handling: */
  opt_handling_res = 
    bmc_CommandsWithout_k_l_OptionsHandling(argc, argv, &invarprop, Prop_Invar,
					    &bCreateDimacs);

  if (opt_handling_res == SUCCESS_REQUIRED_HELP) {
    BMC_CMD_RESTORE_OPTIONS(); 
    return UsageBmcCheckInvar();
  }
  if (opt_handling_res != SUCCESS) {
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }
  /* ----------------------------------------------------------------------- */


  /* make sure the model hierarchy has been flattened */
  if (cmp_struct_get_flatten_hrc (cmps) == 0) {
    fprintf (nusmv_stderr, "The hierarchy must be flattened before. Use "\
	     "the \"flatten_hierarchy\" command.\n");
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }

  /* make sure bmc has been set up */
  if (cmp_struct_get_bmc_setup(cmps) == 0) {
    fprintf (nusmv_stderr, "Bmc must be setup before. Use "\
	     "the \"bmc_setup\" command.\n");
    BMC_CMD_RESTORE_OPTIONS(); 
    return 1;
  }

  /* prepare the list of properties if no property was selected: */
  if (invarprop == NIL_PTR(Prop_Ptr)) {
    lsList props = Prop_Db_GetPropsOfType(Prop_Invar);
    lsGen  iterator; 
    Prop_Ptr prop;

    nusmv_assert(props != LS_NIL);

    lsForEachItem(props, iterator, prop) {
      Bmc_GenSolveInvar((Prop_Ptr)prop, bCreateDimacs, TRUE);
    }

    lsDestroy(props, NULL); /* the list is no longer needed */
  }
  else {
    /* its time to generate dimacs: */
    Bmc_GenSolveInvar(invarprop, bCreateDimacs, TRUE);
  }

  BMC_CMD_RESTORE_OPTIONS(); 
  return 0; 
}

/**Function********************************************************************

  Synopsis           [Usage string for command check_invar_bmc]

  Description        []

  SideEffects        [None]

  SeeAlso            [CommandCheckInvarBmc]

******************************************************************************/
static int UsageBmcCheckInvar(void)
{
  fprintf(nusmv_stderr, "\nUsage: check_invar_bmc [-h | -n idx | -p \"formula\"] [-o filename]\n");
  fprintf(nusmv_stderr, "  -h \t\tPrints the command usage\n");
  fprintf(nusmv_stderr, "  -n idx\tChecks the INVAR property specified with <idx>\n");
  fprintf(nusmv_stderr, "  -p \"formula\"\tChecks the specified INVAR propositional property\n");
  fprintf(nusmv_stderr, "\t\tIf no property is specified, checks all INVAR properties.\n"); 
  fprintf(nusmv_stderr, "  -o filename\tDumps the invar to dimacs file named <filename>. For <filename> patterns are allowed.\n\n");

  return(1);
}

/**Function********************************************************************

  Synopsis           [Bmc commands options handling for commands with 
  -k and -l options]

  Description        [Handles options of bmc commands 
  {check,gen}_ltlspec_bmc_*. pProperty will contain the property must
  be checked out, or NULL if all properties must be checked. Initial value
  of ltlspec *must* be NULL<BR>
  This function will check the values specified for length and loopback are 
  compatible with their constrains. <BR>
  dimacsSpecified will be put to one (1) if the -o option is specified, zero 
  (0) otherwise. <BR>
  Returns GENERIC_ERROR if an error has occurred;
  Returns SUCCESS_REQUIRED_HELP if -h options had been specified; 
  Returns SUCCESS in all other cases. 
  ]

  SideEffects        [High level options (system environment variables) might
  change: make sure to save options before to call this function. pProperty 
  might change if one or more valid properties are specified. dimacsSpecified
  will change if -o option is specified]

  SeeAlso            []

******************************************************************************/
static outcome bmc_CommandsWith_k_l_OptionsHandling(int argc, char ** argv, 
						    Prop_Ptr* pProperty,
						    Prop_Type type, 
						    int* dimacsSpecified) 
{
  int c;
  int prop_idx;

  /* to keep requested loopback string: if were specified, option -l should 
   be evaluated after -k option is evaluated */
  char* szLoopback = NIL(char); 

  nusmv_assert(*pProperty == NIL_PTR(Prop_Ptr)); /* precondition contract */  

  *dimacsSpecified = 0;

  util_getopt_reset();
  while((c = util_getopt(argc, argv, "hn:p:k:l:o:")) != EOF) {
    switch (c) {
    case 'h':
      return SUCCESS_REQUIRED_HELP;

    case 'n':
      {
	char *strPropIdx = NIL(char);
	
	/* check if a formula has been already specified: */
	if (*pProperty != NIL_PTR(Prop_Ptr)) {
	  bmcErrorPropertyAlreadySpecified();
	  return GENERIC_ERROR;
	}
	
	strPropIdx = util_strsav(util_optarg);

	/* check if property idx is ok */
	prop_idx = Prop_GetPropIdxFromString(strPropIdx);
	if (prop_idx == -1) {
	  /* error messages have been already shown */
	  return GENERIC_ERROR;
	}
	  
	/* here property idx is ok */
	*pProperty = Prop_Db_GetNum(prop_idx);
	if ( Prop_CheckType(*pProperty, type) != 0 ) {
	  return GENERIC_ERROR;
	}
	
	break;
      } /* case 'n' */

    case 'p':
      {
	char *strFormula = util_strsav(util_optarg);
	int  idx = -1;
	/* check if a formula has been already specified: */
	if (*pProperty != NIL_PTR(Prop_Ptr)) {
	  bmcErrorPropertyAlreadySpecified();
	  return GENERIC_ERROR;
	}
	idx = Prop_Db_PropParseAndAdd(strFormula, type);
	if (idx == -1) {
	  bmcErrorInvalidProperty(strFormula);
	  return GENERIC_ERROR;
	}
	else {
	  /* index is ok */
	  *pProperty = Prop_Db_GetNum(idx);
	}
	break;
      }
	 
    case 'k':
      {
	char *strNumber = util_strsav(util_optarg);
	int k;
	if (util_str2int(strNumber, &k) != 0) {
	  bmcErrorInvalidNumber(strNumber);
	  return GENERIC_ERROR;
	}
	  
	/* k and l constrains are going to be check out of switch statement */ 
	set_bmc_pb_length(options, k); 
	break;  
      }
       

    case 'l':
      {
	szLoopback = util_strsav(util_optarg);
	/* checking of loopback value is moved after command line processing
	   to allow any -k option evaluation before (see following call to
	   the bmcSetLoopbackAsString function)
	*/
	break;
      }
      
    case 'o':
      {
	char* szFilename = util_strsav(util_optarg);
	set_bmc_dimacs_filename(options, szFilename);
	*dimacsSpecified = 1;
	break;
      }	    

    default:
      return GENERIC_ERROR;
    } /* switch case */
  } /* end of cmd line processing */
  
  /* checks if there are unexpected options: */
  if (argc != util_optind) {
    fprintf(nusmv_stderr, "You specified one or more invalid options.\n\n");
    return GENERIC_ERROR;
  }  

  /* ------------------------------------------------------------------------*/
  /* Checking of k,l constrains:                                             */
  if (szLoopback != NIL(char)) {
    if (bmcSetLoopbackAsString(szLoopback) != SUCCESS) {
      bmcErrorInvalidLoopback(szLoopback);
      return GENERIC_ERROR;
    }
  }
  if (bmcCheck_k_l() != SUCCESS) {
    bmcErrorInvalid_k_l(get_bmc_pb_length(options), 
			get_bmc_pb_loop_user_pov(options));
    return GENERIC_ERROR;
  }
  /* ------------------------------------------------------------------------*/
  
  return SUCCESS;
}



/**Function********************************************************************

  Synopsis           [Bmc commands options handling for commands without 
  -k and -l options]

  Description        [Handles options of bmc commands 
  {check,gen}_invar_bmc_*. pProperty will contain the property must
  be checked out, or NULL if all properties must be checked. Initial value
  of ltlspec *must* be NULL<BR>
  dimacsSpecified will be put to one (1) if the -o option is specified, zero 
  (0) otherwise. <BR>
  Returns GENERIC_ERROR if an error has occurred;
  Returns SUCCESS_REQUIRED_HELP if -h options had been specified; 
  Returns SUCCESS in all other cases. 
  ]

  SideEffects        [High level options (system environment variables) might
  change: make sure to save options before to call this function. pProperty 
  might change if one or more valid properties are specified. dimacsSpecified
  will change if -o option is specified]

  SeeAlso            []

******************************************************************************/
static outcome bmc_CommandsWithout_k_l_OptionsHandling(int argc, char ** argv, 
						       Prop_Ptr* pProperty, 
						       Prop_Type type, 
						       int* dimacsSpecified) 
{
  int c;
  int prop_idx;

  nusmv_assert(*pProperty == NIL_PTR(Prop_Ptr)); /* precondition contract */  

  *dimacsSpecified = 0;

  util_getopt_reset();
  while((c = util_getopt(argc, argv, "hn:p:o:")) != EOF) {
    switch (c) {
    case 'h':
      return SUCCESS_REQUIRED_HELP;

    case 'n':
      {
	char *strPropIdx = NIL(char);
	
	/* check if a formula has been already specified: */
	if (*pProperty != NIL_PTR(Prop_Ptr)) {
	  bmcErrorPropertyAlreadySpecified();
	  return GENERIC_ERROR;
	}
	
	strPropIdx = util_strsav(util_optarg);

	/* check if property idx is ok */
	prop_idx = Prop_GetPropIdxFromString(strPropIdx);
	if (prop_idx == -1) {
	  /* error messages have been already shown */
	  return GENERIC_ERROR;
	}
	  
	/* here property idx is ok */
	*pProperty = Prop_Db_GetNum(prop_idx);
	if ( Prop_CheckType(*pProperty, type) != 0 ) {
	  return GENERIC_ERROR;
	}
	
	break;
      } /* case 'n' */

    case 'p':
      {
	char *strFormula = util_strsav(util_optarg);
	int idx;

	/* check if a formula has been already specified: */
	if (*pProperty != NIL_PTR(Prop_Ptr)) {
	  bmcErrorPropertyAlreadySpecified();
	  return GENERIC_ERROR;
	}

	idx = Prop_Db_PropParseAndAdd(strFormula, type); 

	if (idx == -1) {
	  bmcErrorInvalidProperty(strFormula);
	  return GENERIC_ERROR;
	}
	else {
	  /* index is ok */
	  *pProperty = Prop_Db_GetNum(idx);
	}
	break;
      }
	 
    case 'o':
      {
	char* szFilename = util_strsav(util_optarg);
	set_bmc_invar_dimacs_filename(options, szFilename);
	*dimacsSpecified = 1;
	break;
      }	    

    default:
      return GENERIC_ERROR;
    } /* switch case */
  } /* end of cmd line processing */

  /* checks if there are unexpected options: */
  if (argc != util_optind) {
    fprintf(nusmv_stderr, "You specified one or more invalid options.\n\n");
    return GENERIC_ERROR;
  }  
    
  return SUCCESS;
}



static outcome bmcSetLoopbackAsString(const char* str) 
{
  outcome ret;
  int l;

  ret = Bmc_ConvertLoopbackFromString(str, &l);
  if (ret == SUCCESS)  set_bmc_pb_loop(options, l);

  return ret;
}

/**Function********************************************************************

  Synopsis           [Stores current vaues of length and loopback into two
  given variables]

  Description        [Use bmcRestore_k_l to restore stored values into 
  length and loopback]

  SideEffects        [k_store and l_store addressed locations will change]

  SeeAlso            [bmcRestore_k_l]

******************************************************************************/
static void bmcSave_k_l(int* k_store, int* l_store)
{
  *k_store = get_bmc_pb_length(options);
  *l_store = get_bmc_pb_loop_user_pov(options);
}

static void bmcRestore_k_l(const int k_store, const int l_store)
{
  set_bmc_pb_length(options, k_store);
  set_bmc_pb_loop(options, l_store);
}

static outcome bmcCheck_k_l()
{
  outcome ret = GENERIC_ERROR;
  int k, l;

  k = get_bmc_pb_length(options);
  l = get_bmc_pb_loop(options);
  if ( (k>=0) && (l<=k) &&
       ((l==BMC_NO_LOOP) || (l==BMC_ALL_LOOPS) || (l>=0)) )  {
    ret = SUCCESS;
  }

  return ret;
}
  



static void bmcErrorInvalidProperty(const char* szFormula)
{
  if (szFormula != NIL(char)) {
    fprintf(nusmv_stderr, "Error: property \"%s\" is not valid\n", szFormula);
  }
  else {
    fprintf(nusmv_stderr, "Error: you given an empty property\n");
  }
}


static void bmcErrorPropertyAlreadySpecified()
{
  fprintf(nusmv_stderr, "Error: It is not possible to specify multiple properties to be checked \n");
  fprintf(nusmv_stderr, "      \"using -p <formula>\" and/or \"-n <property_idx>\" options\n\n"); 
}

static void bmcErrorInvalidNumber(const char* szNumber)
{
  if (szNumber != NIL(char)) {
    fprintf(nusmv_stderr, "Error: string \"%s\" is not valid. An integer was expected.\n", szNumber);
  }
  else {
    fprintf(nusmv_stderr, "Error: you given an empty string when integer was expected\n");
  }
}

static void bmcErrorInvalidLoopback(const char* szLoopbackString) 
{
  if (szLoopbackString != NIL(char)) {
    fprintf(nusmv_stderr, 
	    "Error: string \"%s\" is not a valid loopback value.\n", 
	    szLoopbackString);
  }
  else {
    fprintf(nusmv_stderr, 
	    "Error: you given an empty string as loopback value\n");
  }
}

static void bmcErrorInvalid_k_l(const int k, const int l)
{
  char buf[16];
  
  Bmc_ConvertLoopbackFromInteger(l, buf, sizeof(buf));
  fprintf(nusmv_stderr, 
	  "Error: length=%d and loopback=%s are not compatible.\n",
	  k, buf);
}

/**Function********************************************************************

  Synopsis           [Stores current default dimacs filenames into a buffer]

  Description        [This function allocates as space as needed to contain
  current default dimacs filename string. The given pointer will point to
  the just allocated buffer. It is strongly suggested you use function 
  bmcRestoreDimacsFilename to dealloc the buffer and to restore the original
  dimacs filename value]

  SideEffects        [szFilenameDimacs and szFilenameInvarDimacs will change]

  SeeAlso            [bmcRestoreDimacsFilenames]

******************************************************************************/
static void 
bmcSaveDimacsFilenames(char** szFilenameDimacs, char** szFilenameInvarDimacs)
{
  char* szDimacs = get_bmc_dimacs_filename(options);
  char* szInvar  = get_bmc_invar_dimacs_filename(options);

  *szFilenameDimacs = ALLOC(char, strlen(szDimacs)+sizeof(char));
  nusmv_assert(*szFilenameDimacs != NIL(char));
  strcpy(*szFilenameDimacs, szDimacs);

  *szFilenameInvarDimacs = ALLOC(char, strlen(szInvar)+sizeof(char));
  nusmv_assert(*szFilenameInvarDimacs != NIL(char));
  strcpy(*szFilenameInvarDimacs, szInvar);
}


/**Function********************************************************************

  Synopsis           [Restores the original value of dimacs filenames option]

  Description        [Use this function to restore dimacs filenames saved using
  bmcSaveDimacsFilenames]

  SideEffects        [szFilenameDimacs and szFilenameInvarDimacs will be 
  deleted]

  SeeAlso            [bmcSaveDimacsFilenames]

******************************************************************************/
static void 
bmcRestoreDimacsFilenames(char* szFilenameDimacs, char* szFilenameInvarDimacs)
{
  assert(szFilenameDimacs != NIL(char));
  assert(szFilenameInvarDimacs != NIL(char));

  set_bmc_dimacs_filename(options, szFilenameDimacs);
  set_bmc_invar_dimacs_filename(options, szFilenameInvarDimacs);

  FREE(szFilenameDimacs);
  FREE(szFilenameInvarDimacs);
}

