/**CFile***********************************************************************

  FileName    [bmcSatsolver.c]

  PackageName [bmc]

  Synopsis    [BMC interface to SAT solvers.]

  Description [Contains the primitives for dumping a CNF formula in a
  DIMACS file, calling an external solver (e.g. SATO), and calling an
  internal solver (e.g. SIM).]

  SeeAlso     []

  Author      [Alessandro Cimatti and Lorenzo Delana]

  Copyright   [
  This file is part of the ``bmc'' package of NuSMV version 2. 
  Copyright (C) 2000-2001 by ITC-irst and University of Trento. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "bmcInt.h"


/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static void CnfWriteDimacs ARGS((FILE * outFile, lsList clauses, 
				 lsList vars, int maxVar));

static sat_result CallSimSolver ARGS((lsList clauses, lsList vars, int maxVar,
				      int * parms, lsList lsResult));

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis    [Writes a BMC problem in Dimacs format.]

  Description [The Dimacs file is annotated with a description of the
  model checking problem, and with the mapping from model variables to
  dimacs variables
  Returns FILE_ERROR if an error had occured, SUCCESS otherwise]

  SideEffects [none]

  SeeAlso     [] 

******************************************************************************/

outcome Bmc_DimacsWriteProblem (char * fname, int k,
				lsList CnfClauses, lsList CnfVars,
				int CnfMaxVar)
{
  FILE * dimacsfile;
  int time, varindex;

  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "\nDumping CNF on DIMACS file %s ...", fname);
  }
  dimacsfile = fopen(fname, "w");

  if (dimacsfile==NULL) {
    /* error opening file */
    return FILE_ERROR; 
  }
  
  fprintf(dimacsfile, "c BMC problem generated by NuSMV\n");
  fprintf(dimacsfile, "c Time steps from 0 to %d, %d State Variables\n",
	  k, Bmc_VarsMgrGetNum());

  fprintf(dimacsfile, "c Model to Dimacs Conversion Table\n");
  for(time = 0; time <= k; time++) {
    fprintf(dimacsfile, "c \nc @@@@@ Time %d\n", time);
    for(varindex = 0; varindex < Bmc_VarsMgrGetNum(); varindex++) {
      fprintf(dimacsfile, "c CNF variable %d => Time %d, Model Variable ",
	      ((time + 2) * Bmc_VarsMgrGetNum()) + varindex, time);
      print_node(dimacsfile, Bmc_GetIthStatevar(varindex));
      fprintf(dimacsfile, "\n");
    }
  }
  fprintf(dimacsfile, "c \n");

  CnfWriteDimacs(dimacsfile, CnfClauses, CnfVars, CnfMaxVar);

  if (fclose(dimacsfile) == EOF) {
    return FILE_ERROR;
  }

  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, " done!\n");
  }
  return SUCCESS;
}


/**Function********************************************************************

  Synopsis           [Dumps Davinci format to a specified file]

  Description        [The filename must not contain extension, which will 
  be appended from this function. Returns SUCCESS if the dump has been 
  performed successfully, otherwise returns FILE_ERROR]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
outcome Bmc_DavinciWriteProblem(rbc_ptr rbc, char* fname)
{
  static const char FILE_EXT[] = ".davinci";
  char pathfname [4096 + 1];
  FILE * f;

  nusmv_assert( strlen(fname) < (sizeof(pathfname) - sizeof(FILE_EXT)) );

  strcpy (pathfname, fname);
  strcat (pathfname, FILE_EXT);

  f = fopen (pathfname, "w");
  if (f == NULL) {
    /* error opening file */
    return FILE_ERROR; 
  }
   
  Bmc_RbcDumpDavinci(Bmc_VarsMgrGetRbcMgr(), rbc, f);

  if (fclose(f) == EOF) {
    return FILE_ERROR;
  }
  
  return SUCCESS;
}


/**Function********************************************************************

  Synopsis    [Bmc interface to Sim as internal solver]

  Description [
  If the sim solver successfully returns a model, the representation of that
  model can be found into the list lsResult, which must be previously created.
  The calle function has the responsibility for the list allocation and 
  deallocation.   
  Returns:
               <ul>
               <li> SAT_INTERNAL_ERROR         if an error occurred
               <li> SAT_UNSATISFIABLE_PROBLEM  if the formula is unsatisfiable
               <li> SAT_SATISFIABLE_PROBLEM    if the assignment is satisfying 
	       the formula. In this case lsResult points to a list of literals
               </ul>]

  SideEffects [the list lsResult could be changed]

  SeeAlso     [CallSimSolver] 

******************************************************************************/
sat_result Bmc_CallSimSolver(lsList CnfClauses, lsList CnfVars, int CnfMaxVar, 
			     lsList lsResult)
{
  static int* p = NULL;
  long start; /* start time */
  sat_result result; 

  /* this code is waiting for Sim_ParamQuit() becomes available */
  if (p == NULL) {
    p = Sim_ParamInit(); /* sim parameters */
    nusmv_assert(p != NULL);
  }

  /* This is essential: correct number of variables and clauses! */
  p = Sim_ParamSet(p, SIM_MAX_VAR_NUM, CnfMaxVar);
  p = Sim_ParamSet(p, SIM_MAX_CL_NUM, lsLength(CnfClauses));
  
  /* A few additional settings: 
     - using UNITIE heuristic (another good one for BMC is SIM_SATZ_HEUR)
     - using a TRACE on search tree nodes and leafs with step 10 (0 no trace)
     - using independent variables for splitting (0 all variables)
  */
  p = Sim_ParamSet(p, SIM_HEURISTICS, SIM_UNITIE_HEUR);
  p = Sim_ParamSet(p, SIM_RUN_TRACE, 0);
  p = Sim_ParamSet(p, SIM_INDEP_PROPS, 1);

  /* Call SIM */
  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "\nCalling SIM\n");
  }
  
  start = util_cpu_time();
  result = CallSimSolver(CnfClauses, CnfVars, CnfMaxVar, p, lsResult);
  
  if (opt_verbose_level_gt(options, 1)) {
    fprintf(stderr, "The time spent in SIM is: %2.3f secs\n",
	    (util_cpu_time() - start)/1000.0);
  }
  return result;
}



/**Function********************************************************************

  Synopsis    [Interface to external solver SATO.]

  Description [Interface to external solver Sato.
               Returns:
               <ul>
               <li> SAT_INTERNAL_ERROR        if an error occurred
               <li> SAT_UNSATISFIABLE_PROBLEM if the formula is unsatisfiable
               <li> SAT_SATISFIABLE_PROBLEM   if the formula is unsatisfiable
               </ul>]     

  SideEffects [none]

  SeeAlso     [] 

******************************************************************************/

sat_result Bmc_CallSatoSolver (char * fname)
{
  char* command_string;
  const char* sato_path = "/tmp/sato";
  const char* szTemplate= "time %s %s >& %s.sato.out";
  time_t start_time = 0;
  int result = 0;
  sat_result ret = SAT_UNAVAILABLE;

  command_string = ALLOC( char, (strlen(szTemplate) +
				 strlen(sato_path) +
				 strlen(fname) +
				 strlen(fname)) );
  nusmv_assert(command_string != NULL);

  /* Generate command string */
  sprintf(command_string, szTemplate,
	  sato_path, fname, fname);

  /* The call */
  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "Calling sato as \"%s\"\n", command_string);
  }
  start_time = time((time_t *)NULL);
  result = system(command_string);
  if (opt_verbose_level_gt(options, 1)) {
    fprintf(stderr,
	    "Time spent in the call to SATO: %2.3d secs\n",
	    (int)( (time((time_t *)NULL) - start_time) ) 
	   );
  }

  if (result < 0) {
    fprintf(nusmv_stdout, "**** SATO: problems at runtime\n");
    ret = SAT_INTERNAL_ERROR;
  }
  else {
    /* Retrieve run time reported by sato */
    sprintf(command_string,
	    "grep \"run time (seconds)\" %s.sato.out",
	    fname);
    result = system(command_string);

    /* Check for satisfiability - a bit approximate */
    sprintf(command_string,
	    "grep \"The number of found models is \" %s.sato.out",
	    fname);
    result = system(command_string);

    if (result == 0) {
      fprintf(nusmv_stdout, "**** SATO: Model found!\n");
      ret = SAT_SATISFIABLE_PROBLEM;
    }
    else {
      fprintf(nusmv_stdout, "**** SATO: Unsatisfiable problem!\n");
      ret = SAT_UNSATISFIABLE_PROBLEM;
    }
  }

  FREE(command_string);
  return ret;
}




/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/



/**Function********************************************************************

  Synopsis [Ouput a list of clauses on a DIMACS file.]

  Description [Ouput a list of clauses as a DIMACS file. The input
  file is supposed to be open and closed out of this function.]

  SideEffects [none]

  SeeAlso     [] 

******************************************************************************/
static void CnfWriteDimacs(
  FILE * outFile, 
  lsList clauses, 
  lsList vars, 
  int maxVar)
{
  lsGen  genCl, genLit;
  lsList cl;
  int    lit;
  
  /* checks file has been opened successfully */
  nusmv_assert(outFile != NULL);

  /* Stamp the beginning of file processing. */
  fprintf(outFile, "c Begin of CnfWriteDimacs\n");

  /* Print the model variables as a "special" comment line. */
  fprintf(outFile, "c model %d\n", lsLength(vars));
  fprintf(outFile, "c ");
  genLit = lsStart(vars);
  while (lsNext(genLit, (lsGeneric*) &lit, LS_NH) == LS_OK) {
    fprintf(outFile, "%d ", lit);
  }
  lsFinish(genLit);
  fprintf(outFile, "0\n");

  /* Print the problem header. */
  fprintf(outFile, "p cnf %d %d\n", maxVar, lsLength(clauses));

  /* Print the clauses. */
  genCl = lsStart(clauses);
  while (lsNext(genCl, (lsGeneric*) &cl, LS_NH) == LS_OK) {
    genLit = lsStart(cl);
    while (lsNext(genLit, (lsGeneric*) &lit, LS_NH) == LS_OK) {
      fprintf(outFile, "%d ", lit);
    }
    lsFinish(genLit);
    fprintf(outFile, "0\n");
  }
  lsFinish(genCl);

  /* Stamp the beginning of file processing. */
  fprintf(outFile, "c End of CnfWriteDimacs\n");

} /* End of CnfWriteDimacs. */




/**Function********************************************************************

  Synopsis    [A simple function to interface a call to Sim.
  If the sim solver successfully returns a model, the representation of that
  model can be found into the list lsResult, which must be previously created. 
               Returns:
               <ul>
               <li> SAT_INTERNAL_ERROR         if an error occurred
               <li> SAT_UNSATISFIABLE_PROBLEM  if the formula is unsatisfiable
               <li> SAT_SATISFIABLE_PROBLEM    if the assignment satisfying 
	       the formula. In this case lsResult points to a list of literals
               </ul>]

  Description []

  SideEffects [the list lsResult could be changed]

  SeeAlso     [] 

******************************************************************************/

static sat_result CallSimSolver(
  lsList clauses, 
  lsList vars, 
  int maxVar,
  int * parms, 
  lsList lsResult)
{
  lsGen    genCl, genLit;
  lsList   cl;
  sat_result result;
  int    * m;
  int      lit, res, i, clId;

  /* Initialize Sim. */
  Sim_DllInit(parms);

  /* Upload the clauses. */
  genCl = lsStart(clauses);
  while (lsNext(genCl, (lsGeneric*) &cl, LS_NH) == LS_OK) {
    genLit = lsStart(cl);
    clId = Sim_DllNewCl();
    while (lsNext(genLit, (lsGeneric*) &lit, LS_NH) == LS_OK) {
      res = Sim_DllAddLit(clId, lit);
      switch (res) {
      case 0 : 
	/* A tautology: go to the end of the clause. */
	while (lsNext(genLit, (lsGeneric*) &lit, LS_NH) == LS_OK);
	/* Initiate a new clause. */
	clId = Sim_DllNewCl();
	break;
      case -1 :
	/* The program is gone berserk... */
	return SAT_INTERNAL_ERROR;
      default :
	/* Everything is Ok. */
	break;
      }
    }
    lsFinish(genLit);
    /* Commit the clause. */
    Sim_DllCommitCl(clId);
  }
  lsFinish(genCl);

  /* Upload the independent variables. */
  genLit = lsStart(vars);
  while (lsNext(genLit, (lsGeneric*) &lit, LS_NH) == LS_OK) {
    Sim_DllPropMakeIndep(lit);
  }
  lsFinish(genLit);


  /* Have Sim compute a solution. */

  if (Sim_DllSolve() == SIM_SAT) {
    m = Sim_DllGetStack();
    for (i = 1; i < m[0]; i++) {
      (void) lsNewEnd(lsResult, (lsGeneric)m[i], LS_NH);
    }
    result = SAT_SATISFIABLE_PROBLEM; 
  } 
  else {
    result = SAT_UNSATISFIABLE_PROBLEM; 
  }    

  /* Clean up SIM. */
  Sim_DllClear();
  
  return result;

} /* End of CallSim. */


