/**CFile***********************************************************************

  FileName    [bmcVarsMgr.c]

  PackageName [bmc]

  Synopsis    [Variables Manager for the <tt>bmc</tt> package]

  Description []

  SeeAlso     []

  Author      [Alessandro Cimatti and Lorenzo Delana]

  Copyright   [
  This file is part of the ``bmc'' package of NuSMV version 2. 
  Copyright (C) 2000-2001 by ITC-irst and University of Trento. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "bmcInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/




/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/
/**Struct**********************************************************************

  Synopsis    [Variables Manager (core data structure)]

  Description [Variable Manager is a private structure owned by this module.
  In order to access to its members use the supplied methods, which give you
  access to the general part of the structure tied to rbc generic interface. 
  The structural part specific of the current implementation (tables and other
  specific sub-structures) is not accessible from the outer space.]

  SeeAlso     [bmc_vars_init, bmc_vars_quit]

******************************************************************************/
struct bmc_vars_mgr_TAG
{
  /* ----------------------------------------------------------------------- */
  /*    Accessible generic part, tied to rbc interface                       */
  /* ----------------------------------------------------------------------- */
  RbcManager_ptr env;              /* the variables environment */
  int num;                         /* the number of state variables */
  int maxtime;                     /* max time reached until now for the
				       variables environment */

  /* ----------------------------------------------------------------------- */
  /*    Private part specific of implementation                              */
  /* ----------------------------------------------------------------------- */
  hash_ptr statevar2curvar_table;   /* table for statevar -> curvar
				       conversion */
  node_ptr* curvar2statevar_table; /* table for curvar -> statevar
				       conversion */
  int* cnexp_subst_array;          /* array for current, next vars 
				       substitution */
};


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/
typedef struct bmc_vars_mgr_TAG bmc_vars_mgr;


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/**Variable********************************************************************

  Synopsis    [Variables Manager (instanced container variable)]

  Description [This global variable is private to this module]

  SeeAlso     [bmc_vars_init, bmc_vars_quit]

******************************************************************************/
static bmc_vars_mgr* vars_mgr = (bmc_vars_mgr*)NULL;

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/


/**AutomaticEnd***************************************************************/



/*---------------------------------------------------------------------------*/
/* Declaration of private functions                                          */
/*---------------------------------------------------------------------------*/

/* Following private wrappers assure a more generic access to the rbc part 
   into Variables Manager structure: */ 
static void bmc_vars_mgr_set_env     ARGS((RbcManager_ptr pNewRbcManager));
static void bmc_vars_mgr_set_num     ARGS((int iNewNum));
static void bmc_vars_mgr_set_maxtime ARGS((int iMaxTime));
static void scan_state_vars_bexp     ARGS((node_ptr state_vars_bexp));


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis           [<b>Returns</b> the
  <tt>number of variables currently handled by the Variables Manager</tt>]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
int Bmc_VarsMgrGetNum(void)
{
  return vars_mgr->num;
}


/**Function********************************************************************

  Synopsis           [<b>Returns</b> the
  <tt>maximum time currently handled by the variable environment</tt>]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
int Bmc_VarsMgrGetMaxtime(void)
{
  return vars_mgr->maxtime;
}


/**Function********************************************************************

  Synopsis           [<b>Returns</b> the
  <tt>variable environment data structure</tt>]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
RbcManager_ptr Bmc_VarsMgrGetRbcMgr(void)
{
  return vars_mgr->env;
}


/**Function********************************************************************

  Synopsis           [<b>Converts</b> given <b>boolean expressions list </b> 
  into correspondent <b>reduced boolean circuits list</b>]

  Description        []

  SideEffects        [rbc hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr Bmc_MapBexp2Rbc (node_ptr bexp_list)
{
  if (bexp_list == Nil) 
  {
    return(Nil);
  }
  else 
  {
    return cons( (node_ptr)Bmc_Bexp2Rbc(car(bexp_list)),
		 Bmc_MapBexp2Rbc(cdr(bexp_list)) );
  }
}


/**Function********************************************************************

  Synopsis           [<b>Converts</b> given <b>boolean expression</b> into
  correspondent <b>reduced boolean circuit</b>]

  Description        []

  SideEffects        [rbc hash may change]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_Bexp2Rbc (node_ptr bexp)
{
  /* if given expression is Nil, return truth rbc */
  /* RC: Nil value can be used in AND sequences, so a true value must
     be returned */
  if (bexp == Nil) {
    return Bmc_RbcTruth (Bmc_VarsMgrGetRbcMgr());
  }

  switch (node_get_type (bexp)) {
  case NUMBER:                          /* NUMBER   */
    {
      int arg = (int) car (bexp);

      /* A NUMBER must be either 0 (falsity) or 1 (truth) */
      nusmv_assert ((arg == 0) || (arg == 1));

      if (arg == 0) 
	return Bmc_RbcTruth(Bmc_VarsMgrGetRbcMgr());
      else     
	return Bmc_RbcFalsity(Bmc_VarsMgrGetRbcMgr());
    }

  case TRUEEXP:                         /* TRUEEXP  */
    return Bmc_RbcTruth(Bmc_VarsMgrGetRbcMgr());

  case FALSEEXP:                        /* FALSEEXP */
    return Bmc_RbcFalsity(Bmc_VarsMgrGetRbcMgr());

  case NEXT:                            /* NEXT     */
    /* converts NEXT arg (a state variable) into correspondent next variable
       and return this rbc variable */
    return Bmc_Statevar2Nextvar (car (bexp));

  case NOT:                             /* NOT      */
    /* NOT arg is converted into an rbc, and its negation is returned */
    return Bmc_RbcNot( Bmc_VarsMgrGetRbcMgr(), Bmc_Bexp2Rbc(car(bexp)) );

  case CONS:                            /* CONS TREATED AS AND */
  case AND:                             /* AND      */
    return Bmc_RbcAnd ( Bmc_VarsMgrGetRbcMgr(),  
		     Bmc_Bexp2Rbc (car (bexp)),
		     Bmc_Bexp2Rbc (cdr (bexp)) );

  case OR:                              /* OR       */
    return Bmc_RbcOr ( Bmc_VarsMgrGetRbcMgr(), 
		    Bmc_Bexp2Rbc (car (bexp)),
		    Bmc_Bexp2Rbc (cdr (bexp)) );

  case IFF:                             /* IFF      */
    /* converts IFF args into two rbcs, and return an IFF rbc with converted
       rbcs as childs */
    return Bmc_RbcIff ( Bmc_VarsMgrGetRbcMgr(), 
		     Bmc_Bexp2Rbc (car (bexp)),
		     Bmc_Bexp2Rbc (cdr (bexp)) );

  case IMPLIES:                         /* IMPLIES  */
    /* convert IMPLIES args into two rbcs, and return the IMPLIES rbc with
       converted rbcs as childs */
    return Bmc_RbcImplies ( Bmc_VarsMgrGetRbcMgr(),
			 Bmc_Bexp2Rbc (car (bexp)),
			 Bmc_Bexp2Rbc (cdr (bexp)) );

  case EQUAL:                           /* EQUAL    */
  case ASSIGN:                          /* ASSIGN   */
    /* converts EQUAL and ASSIGN args into two rbcs, and return an IFF rbc
       with converted rbcs as childs */
    return Bmc_RbcIff ( Bmc_VarsMgrGetRbcMgr(),
		     Bmc_Bexp2Rbc (car (bexp)),
		     Bmc_Bexp2Rbc (cdr (bexp)) );
    
  case CASE:                            /* CASE statement  */
    /* converts "if"-"then"-"else" args of CASE into three rbcs, and return
       an rbc If-Then-Else with converted rbcs as childs */
    return Bmc_RbcIte ( Bmc_VarsMgrGetRbcMgr(), 
		     Bmc_Bexp2Rbc (caar (bexp)),
		     Bmc_Bexp2Rbc (cdar (bexp)),
		     Bmc_Bexp2Rbc (cdr (bexp)) );

  case DOT:                              /* VARIABLE */
    return(Bmc_Statevar2Curvar (bexp));

  case ARRAY:                            /* VARIABLE */
  case ATOM:                             /* VARIABLE */
    internal_error("Not DOT node as variable has been found!\n");
    nusmv_assert(false); /* execution should never be here */

  default:                              /* (default action) */
    internal_error("Unaspected case value. Node type = %d\n",
	   node_get_type(bexp));
    nusmv_assert(false); /* execution should never be here */
  }
}


/**Function********************************************************************

  Synopsis           [<b>Extends</b> <tt>maxtime</tt> at given
  <tt>new_maxtime</tt>]

  Description        []

  SideEffects        [rbc hash may change]

  SeeAlso            []

******************************************************************************/
void Bmc_ExtendsMaxtime(int new_maxtime)
{
  nusmv_assert (new_maxtime >= 0);

  /* if is really necessary to do the maxtime extension... */
  if (new_maxtime > Bmc_VarsMgrGetMaxtime()) {
    Bmc_RbcManagerReserve ( Bmc_VarsMgrGetRbcMgr(),
			  /* reserve space for 2 time (curvars, nextvars) +
			   1 time (zero timed vars) +
			   new_maxtime time (timed vars until new_maxtime) */
			  (2 + 1 + new_maxtime) * Bmc_VarsMgrGetNum() );

    /* register maxtime change */
    bmc_vars_mgr_set_maxtime(new_maxtime);
  }
}

/**Function********************************************************************

  Synopsis           [<b>Shift</b> given <i>current</i> <b>expression at 
  next time</b>]

  Description        [Warning: argument 'exp' must contain only variables in 
  current time, otherwise results will be unpredictible, surely inconsistent]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_CurrShiftToNext(rbc_ptr exp)
{
  return Bmc_RbcVarShift ( Bmc_VarsMgrGetRbcMgr(), 
			 exp, 
			 Bmc_VarsMgrGetNum() );
}


/**Function********************************************************************

  Synopsis           [<b>Shift</b> given <i>current, next</i> <b>expression at
  specified <tt>time</tt></b>]

  Description        []

  SideEffects        [extends variables environment and maxtime, if required]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_CurrNextShiftAtTime(rbc_ptr exp, int time)
{
  nusmv_assert (time >= 0);
    
  /* extends maxtime, if required */
  Bmc_ExtendsMaxtime (time + 1);

  /* return exp with variables shifted at given time */
  return Bmc_RbcVarShift ( Bmc_VarsMgrGetRbcMgr(), 
			 exp, 
			 Bmc_Time2Firstvar (time) );
}

/**Function********************************************************************

  Synopsis           [<b>Shift</b> given <i>current, next</i> <b>expression at
  specified <tt>ctime</tt> (current vars at time) and <tt>ntime</tt> (next
  vars at time)</b>]

  Description        []

  SideEffects        [extends variables environment and maxtime, if required]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_CurrNextShiftAtTimes(rbc_ptr exp, int ctime, int ntime)
{
  int i;
  int firstvar_at_ctime, firstvar_at_ntime;
  
  /* lazy evaluation */
  if ( Bmc_RbcIsTrue(Bmc_VarsMgrGetRbcMgr(), exp) ||
       Bmc_RbcIsFalse(Bmc_VarsMgrGetRbcMgr(), exp) ) {
    return exp;
  }

  /* extends maxtime, if required */
  Bmc_ExtendsMaxtime (max (ctime, ntime));

  firstvar_at_ctime = Bmc_Time2Firstvar (ctime);
  firstvar_at_ntime = Bmc_Time2Firstvar (ntime);
  
  for (i = 0;
       i <= (Bmc_VarsMgrGetNum() - 1);
       i++) {
    /* setup substitution for ith-curvar in the array for substitution,
       with ith-timedvar at time ctime */
    vars_mgr -> cnexp_subst_array [i] = firstvar_at_ctime + i;
    
    /* setup substitution for ith-nextvar in the array for substitution,
       with ith-timedvar at time ntime */
    vars_mgr -> cnexp_subst_array [i + Bmc_VarsMgrGetNum()] =
      firstvar_at_ntime + i;
  }

  /* return given exp with symbolic (current, next) variables substituited
     with correspondent timed variables at times ctime and ntime, resp. */
  {
    return Bmc_RbcVarSubst (Bmc_VarsMgrGetRbcMgr(), 
			  exp, 
			  vars_mgr -> cnexp_subst_array);
  }
}

/**Function********************************************************************

  Synopsis           [<b>Make an AND interval</b> of given expression using
  <b>range \[<tt>from</tt>, <tt>to</tt>\]</b>]

  Description        []

  SideEffects        [rbc hash may change]

  SeeAlso            []

******************************************************************************/

rbc_ptr Bmc_MkAndCurrNextInterval(rbc_ptr exp, int from, int to)
{
  if (from > to) {
    /* end of recursion: returns rbc truth */
    return Bmc_RbcTruth(Bmc_VarsMgrGetRbcMgr());
  }
  else {
    rbc_ptr exp_to; 
    nusmv_assert( (from <= to) || (to == -1) );
    exp_to = Bmc_CurrNextShiftAtTime(exp, to);    
    return Bmc_RbcAnd( Bmc_VarsMgrGetRbcMgr(),
		       Bmc_MkAndCurrNextInterval(exp, from, to-1),
		       exp_to 
		     );
  }
}

/**Function********************************************************************

  Synopsis           [<b>Make an OR interval</b> of given expression using
  <b>range \[<tt>from</tt>, <tt>to</tt>\]</b>]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_MkOrCurrNextInterval(rbc_ptr exp, int from, int to)
{
  /* returns given expression in or from time 'from' to time 'to' */
  if (from > to)
    return Bmc_RbcFalsity (Bmc_VarsMgrGetRbcMgr());
  else
    return Bmc_RbcOr ( Bmc_VarsMgrGetRbcMgr(),
		    Bmc_CurrNextShiftAtTime (exp, from),
		    Bmc_MkOrCurrNextInterval(exp, from + 1, to) );
}

/**Function********************************************************************

  Synopsis           [Search the association from given state variable and 
  its relative rbc representation, and returns the rbc form]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_Statevar2Curvar (node_ptr statevar)
{
  rbc_ptr curvar = (rbc_ptr) find_assoc (vars_mgr -> statevar2curvar_table,
					statevar);

  nusmv_assert (curvar != (rbc_ptr) NULL);

  return curvar;
}

/**Function********************************************************************

  Synopsis           [Given the relative index returns the state variable 
  which addresses]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
node_ptr Bmc_GetIthStatevar(int i)
{
  return((vars_mgr -> curvar2statevar_table)[i]);
}

/**Function********************************************************************

  Synopsis           [Given a rbc representation of a variable, returns the 
  relative state variable]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
node_ptr Bmc_Curvar2Statevar (rbc_ptr curvar)
{
  int curvar_index = Bmc_RbcGetVarIndex (Bmc_VarsMgrGetRbcMgr(), curvar);

  nusmv_assert ((curvar_index >= 0) && (curvar_index < Bmc_VarsMgrGetNum()));

  return Bmc_GetIthStatevar (curvar_index);
}

/**Function********************************************************************

  Synopsis           [As Bmc_GetIthStatevar but in terms of rbc 
  representation]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_GetIthCurvar(int i)
{
  nusmv_assert ((i >= 0) && (i < Bmc_VarsMgrGetNum()));

  return Bmc_RbcGetIthVar (Bmc_VarsMgrGetRbcMgr(), i);
}

/**Function********************************************************************

  Synopsis           [As Bmc_GetIthCurvar but relatively to next state 
  variables block]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_GetIthNextvar(int i)
{
  nusmv_assert ((i >= 0) && (i < Bmc_VarsMgrGetNum()));

  return Bmc_RbcGetIthVar ( Bmc_VarsMgrGetRbcMgr(), 
			   Bmc_VarsMgrGetNum() + i );
}

/**Function********************************************************************

  Synopsis           [As Bmc_GetIthNextvar but relatively to 'time' indexed
  state variables block]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_GetIthTimedvar(int i, int time)
{
  nusmv_assert ((i >= 0) && (i < Bmc_VarsMgrGetNum()));

  nusmv_assert (time <= Bmc_VarsMgrGetMaxtime());

  return Bmc_RbcGetIthVar ( Bmc_VarsMgrGetRbcMgr(),
			   (Bmc_VarsMgrGetNum() * (2 + time)) + i );
}

/**Function********************************************************************

  Synopsis           [Given a rbc current variable, returns the rbc variable 
  corrisponding to the next variables block]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_Curvar2Nextvar (rbc_ptr curvar)
{
  int curvar_index = Bmc_RbcGetVarIndex (Bmc_VarsMgrGetRbcMgr(), curvar);

  nusmv_assert ((curvar_index >= 0) && (curvar_index < Bmc_VarsMgrGetNum()));

  return Bmc_GetIthNextvar (curvar_index);
}

/**Function********************************************************************

  Synopsis           [As Bmc_Curvar2Nextvar but current variable must be 
  supplied in state variable form instead of rbc form ]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_Statevar2Nextvar (node_ptr statevar)
{
  return Bmc_Curvar2Nextvar (Bmc_Statevar2Curvar (statevar));
}

/**Function********************************************************************

  Synopsis           [Given a rbc next variable, returns the rbc variable 
  corrisponding to the 'time' index variables block]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_Nextvar2Timedvar (rbc_ptr nextvar, int time)
{
  int nextvar_index = Bmc_RbcGetVarIndex (Bmc_VarsMgrGetRbcMgr(), nextvar);

  nusmv_assert ((nextvar_index >= Bmc_VarsMgrGetNum()) &&
	  (nextvar_index < (2 * Bmc_VarsMgrGetNum())));

  return Bmc_GetIthTimedvar (nextvar_index - Bmc_VarsMgrGetNum(), time);
}

/**Function********************************************************************

  Synopsis           [Given a rbc current variable, returns the rbc variable 
  corrisponding to the 'time' index variables block]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_Curvar2Timedvar (rbc_ptr curvar, int time)
{
  int curvar_index = Bmc_RbcGetVarIndex (Bmc_VarsMgrGetRbcMgr(), curvar);

  nusmv_assert ((curvar_index >= 0) && (curvar_index < Bmc_VarsMgrGetNum()));

  return Bmc_GetIthTimedvar (curvar_index, time);
}

/**Function********************************************************************

  Synopsis           [As Bmc_Curvar2Timedvar but current variable must be 
  supplied in state variable form instead of rbc form]

  Description        []

  SideEffects        [...]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_Statevar2Timedvar (node_ptr statevar, int time)
{
  return Bmc_Curvar2Timedvar (Bmc_Statevar2Curvar (statevar), time);
}

/**Function********************************************************************

  Synopsis           [Given a variable which belongs to a timed block of 
  variables, returns the corrisponding current variable in rbc form]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_Timedvar2Curvar (rbc_ptr timedvar)
{
  int timedvar_index = Bmc_RbcGetVarIndex(Bmc_VarsMgrGetRbcMgr(), timedvar);

  nusmv_assert ((timedvar_index >= (2 * Bmc_VarsMgrGetNum())) &&
	  (timedvar_index < ((2 + 1 + Bmc_VarsMgrGetMaxtime()) *
			     Bmc_VarsMgrGetNum())));

  return Bmc_GetIthCurvar (timedvar_index % Bmc_VarsMgrGetNum());
}

/**Function********************************************************************

  Synopsis           [Given a variable which belongs to a timed block of 
  variables, returns the corrisponding state variable]

  Description        []

  SideEffects        [...]

  SeeAlso            []

******************************************************************************/
node_ptr Bmc_Timedvar2Statevar (rbc_ptr timedvar)
{
  int timedvar_index = Bmc_RbcGetVarIndex(Bmc_VarsMgrGetRbcMgr(), timedvar);

  nusmv_assert ((timedvar_index >= (2 * Bmc_VarsMgrGetNum())) &&
	  (timedvar_index < ((2 + 1 + Bmc_VarsMgrGetMaxtime())
			     * Bmc_VarsMgrGetNum())));

  return Bmc_GetIthStatevar (timedvar_index % Bmc_VarsMgrGetNum());
}

/**Function********************************************************************

  Synopsis           [Given a variable which belongs to the next block of 
  variables, returns the corrisponding state variable]

  Description        []

  SideEffects        [...]

  SeeAlso            []

******************************************************************************/
node_ptr Bmc_Nextvar2Statevar (rbc_ptr nextvar)
{
  int nextvar_index = Bmc_RbcGetVarIndex (Bmc_VarsMgrGetRbcMgr(), nextvar);

  nusmv_assert ((nextvar_index >= Bmc_VarsMgrGetNum()) &&
	  (nextvar_index < (2 * Bmc_VarsMgrGetNum())));
  /* return Bmc_GetIthStatevar (nextvar_index % Bmc_VarsMgrGetNum()); */
  /* RC: substituted from the following version in order to obtain a slightly 
     better performance : */ 
  return Bmc_GetIthStatevar (nextvar_index - Bmc_VarsMgrGetNum());
}

/**Function********************************************************************

  Synopsis           [Given a variable which belongs to a timed block of 
  variables, returns the corrisponding variable in rbc form which belongs to
  the next block of variables]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_Timedvar2Nextvar (rbc_ptr timedvar)
{
  int timedvar_index = Bmc_RbcGetVarIndex(Bmc_VarsMgrGetRbcMgr(), timedvar);

  nusmv_assert ((timedvar_index >= (2 * Bmc_VarsMgrGetNum())) &&
	  (timedvar_index < ((2 + 1 + Bmc_VarsMgrGetMaxtime()) *
			     Bmc_VarsMgrGetNum())));

  return Bmc_GetIthNextvar (timedvar_index % Bmc_VarsMgrGetNum());
}

/**Function********************************************************************

  Synopsis           [Given a variable which belongs to the next block of 
  variables, returns the corrisponding current variable in rbc form]

  Description        []

  SideEffects        [...]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_Nextvar2Curvar (rbc_ptr nextvar)
{
  int nextvar_index = Bmc_RbcGetVarIndex (Bmc_VarsMgrGetRbcMgr(), nextvar);

  nusmv_assert ((nextvar_index >= Bmc_VarsMgrGetNum()) &&
	  (nextvar_index < (2 * Bmc_VarsMgrGetNum())));

  return Bmc_GetIthCurvar (nextvar_index % Bmc_VarsMgrGetNum());
}

/**Function********************************************************************

  Synopsis           [Returns the first variable in rbc form which belongs to
  the given block of variables indexed by 'time']

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
int Bmc_Time2Firstvar(int time)
{
  return ((2 + time) * Bmc_VarsMgrGetNum());
}


/**Function********************************************************************

  Synopsis           [Given a variable index returns the corresponding time]

  Description        [A current variable index returns 0, a next variable 
  returns 1, a timed variable returns its time (p0 returns 0, p1 returns 1, 
  and so on)]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
int Bmc_VarIndex2Time(int varIdx)
{
  int iTime = -1;

  if (varIdx < Bmc_VarsMgrGetNum()) {
    iTime=0; /* current variable */
  }
  else { 
    if (varIdx < (2 * Bmc_VarsMgrGetNum())) {
      iTime = 1; /* next variable */
    }
    else { 
      iTime = (int)(varIdx / Bmc_VarsMgrGetNum()) - 2; /* timed variable */
    }
  }
  
  nusmv_assert( (iTime>=0) && (iTime <= Bmc_VarsMgrGetMaxtime()) );

  return iTime;
}



/**Function********************************************************************

  Synopsis           [Given the index of a timed variable, returns the index
  of the corresponding state variable]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
int Bmc_VarIndex2StateVarIndex(int varIdx) {
  nusmv_assert ( (varIdx >= 0) &&
	   (varIdx < ((2+1+Bmc_VarsMgrGetMaxtime()) * Bmc_VarsMgrGetNum())) );
  
  return varIdx % Bmc_VarsMgrGetNum();
}





/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis           [<b><i>Initialize</i> <tt>BMC</tt> Variables Manager</b>]

  Description        []

  SideEffects        [Alloc <tt>variables manager (data container)</tt><br>
  Follow structures are allocated in the
  <tt>variables manager (data container)</tt>:<br>
  - variables <tt>environment</tt><br>
  - variables <tt>state vars -> current vars</tt> conversion table<br>
  - variables <tt>current vars -> state vars</tt> conversion table<br>
  - variables <tt>substitution array</tt>]

  SeeAlso            [scan_state_vars_bexp]

******************************************************************************/
void BmcVarsMgrInit(node_ptr state_vars_bexp)
{
  /* assert that variables manager module isn't initialized yet */
  nusmv_assert (vars_mgr == ((bmc_vars_mgr *) NULL));

  /* alloc the variables manager core data container */
  vars_mgr = ALLOC (bmc_vars_mgr, 1);
  nusmv_assert(vars_mgr != NULL);

  /* determine the number of state variables */
  bmc_vars_mgr_set_num( llenght (state_vars_bexp) );
  
  if (opt_verbose_level_gt(options, 1)) {
    fprintf(nusmv_stderr, "State variables in RBC manager: %d\n", 
	    Bmc_VarsMgrGetNum()+1);
  }

  /* allocate variables for the two (current and next) frames
     used in initial model construction, and for the zero time frame.
     The variables environment is the manager for the rbc package */
  bmc_vars_mgr_set_env( Bmc_RbcManagerAlloc ((2 + 1) * Bmc_VarsMgrGetNum()) );

  /* setup maxtime indicator in the variables manager */
  bmc_vars_mgr_set_maxtime(0);

  /* allocate the instance for the statevar -> curvar table */
  vars_mgr -> statevar2curvar_table = new_assoc();

  /* allocate the instance for the curvar -> statevar table */
  vars_mgr -> curvar2statevar_table = ALLOC (node_ptr, Bmc_VarsMgrGetNum());
  nusmv_assert(vars_mgr -> curvar2statevar_table != NULL);

  /* scan state variables and update rbc <-> bexp tables */
  scan_state_vars_bexp (state_vars_bexp);

  /* alloc the array for variables substitution */
  if (opt_verbose_level_gt(options, 1)) {
	fprintf( nusmv_stderr, "Allocating cnexp_subst_array: %d locations\n", 
	  2 * (Bmc_VarsMgrGetNum()) );
  }

  vars_mgr -> cnexp_subst_array = ALLOC (int, 2 * Bmc_VarsMgrGetNum());
  nusmv_assert(vars_mgr -> cnexp_subst_array != NULL);
}

/**Function********************************************************************

  Synopsis           [<b><i>Quit</i> from <tt>BMC</tt> Variables Manager</b>]

  Description        []

  SideEffects        [Follow structures are freed in the <tt>variables
  manager (data container)</tt>:<br>
  - variables <tt>substitution array</tt>
  - variables <tt>current vars -> state vars</tt> conversion table<br>
  - variables <tt>state vars -> current vars</tt> conversion table<br>
  - variables <tt>environment</tt><br>
  Free <tt>variables manager (data container)</tt><br>]

  SeeAlso            [bmc_vars_init]

******************************************************************************/
void BmcVarsMgrQuit(void)
{
  if (vars_mgr != ((bmc_vars_mgr *) NULL)) {
    /* free the array for substitution */
    FREE(vars_mgr -> cnexp_subst_array);

    /* free the instance for the curvar -> statevar table */
    FREE(vars_mgr -> curvar2statevar_table);

    /* free the instance for the statevar -> curvar table */
    clear_assoc(vars_mgr -> statevar2curvar_table);

    /* free the variables environment */
    Bmc_RbcManagerFree(Bmc_VarsMgrGetRbcMgr());

    /* free the variables manager core data container */
    FREE(vars_mgr);

    /* set to Nil freed variables manager core data container to allow other
       correct re-initialization of this package */
    vars_mgr = (bmc_vars_mgr * )NULL;
  }
  else {
    /* there was no reason to call this command */
    if (opt_verbose_level_gt(options, 1)) {
      fprintf(nusmv_stderr, 
	      "\nReset of BMC is not necessary (was not initialized yet?)\n");
    }
  }
}





/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [<b>Stores</b> a new <tt>variable environment 
  data structure</tt>]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static void bmc_vars_mgr_set_env (RbcManager_ptr pNewEnv)
{
  vars_mgr->env = pNewEnv;
}


/**Function********************************************************************

  Synopsis           [<b>Sets</b> the
  <tt>number of variables handled by the Variables Manager</tt>]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static void bmc_vars_mgr_set_num (int iNewNum)
{
  vars_mgr->num = iNewNum;
}


/**Function********************************************************************

  Synopsis           [<b>Sets</b> the
  <tt>maximum time currently handled by the variable environment</tt>]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static void bmc_vars_mgr_set_maxtime (int iNewMaxtime)
{
  vars_mgr->maxtime = iNewMaxtime;
}



/**Function********************************************************************

   Synopsis           [Builds all internal structures used in order to 
   perform searches and conversion from different variable representation 
   forms]

   Description        []

   SideEffects        [...]

   SeeAlso            [bmc_vars_init]

******************************************************************************/
static void scan_state_vars_bexp (node_ptr state_vars_bexp)
{
  int i;
  node_ptr scan_state_vars_bexp = state_vars_bexp;
  node_ptr ith_state_var_bexp;

  /* check for the following loop: */
  nusmv_assert( llenght(state_vars_bexp) >= Bmc_VarsMgrGetNum() );

  /* loop over each state variable... */
  for (i = 0; i < Bmc_VarsMgrGetNum(); i++) {
    /* get ith state variable bexp */
    ith_state_var_bexp = car(scan_state_vars_bexp);

    /* update statevar -> curvar table for associate ith state variable with
       ith current variable, that are in the variables environment */
    insert_assoc (vars_mgr -> statevar2curvar_table,
		  ith_state_var_bexp,
		  (node_ptr) Bmc_RbcGetIthVar (Bmc_VarsMgrGetRbcMgr(), i));

    /* update curvar -> statevar table for associate ith current variable,
       that are in the variables environment, with ith state variable */
    vars_mgr -> curvar2statevar_table[i] = ith_state_var_bexp;

    if (opt_verbose_level_gt(options, 1)) {
      fprintf(nusmv_stderr, "[%3d] Boolean variable ", i);
      print_node(nusmv_stderr, car(scan_state_vars_bexp));
      fprintf(nusmv_stderr, "\n");
    }

    /* to the next state vars */
    scan_state_vars_bexp = cdr (scan_state_vars_bexp);
  }
}





