/**CFile***********************************************************************

  FileName    [compileWriteBool.c]

  PackageName [compile]

  Synopsis [Creation of an SMV file containing the flattened
  booleanized model.]

  Description [Creation of an SMV file containing the booleanized and
  flattened model, processes will be removed by explicitly introducing
  a process variable and modifying assignments to take care of
  inertia.]

  SeeAlso     []

  Author      [Marco Roveri]

  Copyright   [
  This file is part of the ``compile'' package of NuSMV version 2. 
  Copyright (C) 2000-2001 by ITC-irst. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "compileInt.h" 

static char rcsid[] UTIL_UNUSED = "$Id: compileWriteBool.c,v 1.5 2001/11/05 10:57:02 none Exp $";

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static int write_flatten_vars ARGS((FILE *, node_ptr, char *));
static int write_flatten_expr ARGS((FILE *, node_ptr, char *));
static int write_flatten_bfexpr ARGS((FILE *, node_ptr, char *));

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/
/**Function********************************************************************

  Synopsis           [Prints the flatten version of an SMV model.]

  Description        [[Prints on the specified file the flatten
  version of an SMV model.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
int Compile_WriteFlattenBool(FILE * out, Fsm_SexpPtr fsm, cmp_struct_ptr s)
{
  fprintf(out, "-- Flattened Boolean model generated from %s\n\n", get_input_file(options)); 
  fprintf(out, "MODULE main\n");

  if (write_flatten_expr(out, Compile_FsmSexpGetInit(fsm), "INIT\n"))
    fprintf(out, "\n");

  if (write_flatten_expr(out, Compile_FsmSexpGetInvar(fsm), "INVAR\n"))
    fprintf(out, "\n");

  if (write_flatten_expr(out, Compile_FsmSexpGetTrans(fsm), "TRANS\n"))
    fprintf(out, "\n");

  if (write_flatten_expr(out, Compile_FsmSexpGetFairness(fsm), "FAIRNESS\n"))
      fprintf(out, "\n\n");

  if (write_flatten_bfexpr(out, cmp_struct_get_spec(s), "SPEC\n"))
      fprintf(out, "\n");

  if (write_flatten_bfexpr(out, cmp_struct_get_compute(s), "COMPUTE\n"))
      fprintf(out, "\n");

  if (write_flatten_bfexpr(out, cmp_struct_get_ltlspec(s), "LTLSPEC\n"))
      fprintf(out, "\n");

  if (write_flatten_bfexpr(out, cmp_struct_get_invar_spec(s), "INVARSPEC\n"))
      fprintf(out, "\n");

  write_flatten_vars(out, input_boolean_variables, "IVAR");
  write_flatten_vars(out, state_boolean_variables, "VAR");

  return 1;
}


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/
/**Function********************************************************************

  Synopsis           [Writes VAR declarations in SMV format on a file.]

  Description        [Writes VAR declarations in SMV format on a file.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static int write_flatten_vars(FILE * out, node_ptr lov, char * str)
{
  if (lov != Nil) fprintf(out, "%s\n", str);
  else return 0;
  while (lov != Nil) {
    node_ptr name = car(lov);
    node_ptr data = lookup_symbol_hash(name);
    
    if (data != Nil && node_get_type(data) == VAR) {
      print_node(out, name);
      if (cdr(data) == boolean_type) {
        fprintf(out, " : boolean;\n");
      } 
      else {
        node_ptr l = cdr(data);

        fprintf(out, " : {");
        while (l != Nil) {
          print_node(out, car(l));
          l = cdr(l);
          if (l != Nil) fprintf(out, ", ");
        }
        fprintf(out, "};\n");
      }
    }
    lov = cdr(lov);
  }
  fprintf(out, "\n");
  return 1;
}

/**Function********************************************************************

  Synopsis           [Writes flattened expression in SMV format on a file.]

  Description        [Writes a generic expression prefixed by a given
  string in SMV format on a file.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static int write_flatten_expr(FILE * out, node_ptr n, char * s)
{
  int res = 0;

  if (n == Nil) return 0;
  switch(node_get_type(n)) {
  case CONS:
  case AND:
    res = write_flatten_expr(out, car(n), s);
    res = write_flatten_expr(out, cdr(n), s);
    break;
  default:
    res = 1;
    if (n != find_node(TRUEEXP, Nil, Nil)) {
      fprintf(out, "%s", s);
      if (node_get_type(n) == CONTEXT) {
      /* We are dealing with a property that has not yet been
         flattened before */
        node_ptr fn = Compile_FlattenSexp(n, Nil);

        print_node(out, fn);
      } else {
        print_node(out, n);
      }
      fprintf(out, "\n\n");
    }
    break;
  }
  return res;
}


/**Function********************************************************************

  Synopsis           [Writes flattened expression in SMV format on a file.]

  Description        [Writes a generic expression prefixed by a given
  string in SMV format on a file.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static int write_flatten_bfexpr(FILE * out, node_ptr n, char * s)
{
  int res = 0;

  if (n == Nil) return 0;
  switch(node_get_type(n)) {
  case CONS:
  case AND:
    res = write_flatten_bfexpr(out, car(n), s);
    res = write_flatten_bfexpr(out, cdr(n), s);
    break;
  default:
    res = 1;
    if (n != find_node(TRUEEXP, Nil, Nil)) {
      fprintf(out, "%s", s);
      if (node_get_type(n) == CONTEXT) {
      /* We are dealing with a property that has not yet been
         flattened and booleanized before */
        node_ptr fn = Compile_FlattenSexp(n, Nil);
        node_ptr bfn = expr2bexpr(fn);
        print_node(out, bfn);
      } else {
        print_node(out, n);
      }
      fprintf(out, "\n\n");
    }
    break;
  }
  return res;
}



